@@ -136,6 +136,17 @@ gst_vaapi_encoder_properties_get_default (const GstVaapiEncoderClass * klass)
    * GstVaapiEncoder:bitrate:
    *
    * The desired bitrate, expressed in kbps.
+   * This is available when rate-control is CBR or VBR.
+   *
+   * CBR: This applies equally to minimum, maximum and target bitrate in the driver.
+   * VBR: This applies to maximum bitrate in the driver.
+   *      Minimum bitrate will be calculated like the following in the driver.
+   *          if (target percentage < 50) minimum bitrate = 0
+   *          else minimum bitrate = maximum bitrate * (2 * target percentage -100) / 100
+   *      Target bitrate will be calculated like the following in the driver.
+   *          target bitrate = maximum bitrate * target percentage / 100
+   *
+   * Note that target percentage is set as 70 currently in GStreamer VA-API.
    */
   GST_VAAPI_ENCODER_PROPERTIES_APPEND (props,
       GST_VAAPI_ENCODER_PROP_BITRATE,
@@ -169,9 +180,86 @@ gst_vaapi_encoder_properties_get_default (const GstVaapiEncoderClass * klass)
           cdata->encoder_tune_get_type (), cdata->default_encoder_tune,
           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
+  /**
+   * GstVaapiEncoder:quality-level:
+   *
+   * The Encoding quality level.
+   */
+  GST_VAAPI_ENCODER_PROPERTIES_APPEND (props,
+      GST_VAAPI_ENCODER_PROP_QUALITY_LEVEL,
+      g_param_spec_uint ("quality-level",
+          "Quality Level", "Encoding Quality Level "
+          "(lower value means higher-quality/slow-encode, "
+          " higher value means lower-quality/fast-encode)",
+          1, 7, 4, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
   return props;
 }
 
+gboolean
+gst_vaapi_encoder_ensure_param_quality_level (GstVaapiEncoder * encoder,
+    GstVaapiEncPicture * picture)
+{
+#if VA_CHECK_VERSION(0,36,0)
+  GstVaapiEncMiscParam *misc;
+
+  /* quality level param is not supported */
+  if (GST_VAAPI_ENCODER_QUALITY_LEVEL (encoder) == 0)
+    return TRUE;
+
+  misc = GST_VAAPI_ENC_QUALITY_LEVEL_MISC_PARAM_NEW (encoder);
+  if (!misc)
+    return FALSE;
+  memcpy (misc->data, &encoder->va_quality_level,
+      sizeof (encoder->va_quality_level));
+  gst_vaapi_enc_picture_add_misc_param (picture, misc);
+  gst_vaapi_codec_object_replace (&misc, NULL);
+#endif
+  return TRUE;
+}
+
+gboolean
+gst_vaapi_encoder_ensure_param_control_rate (GstVaapiEncoder * encoder,
+    GstVaapiEncPicture * picture)
+{
+  GstVaapiEncMiscParam *misc;
+
+  if (GST_VAAPI_ENCODER_RATE_CONTROL (encoder) == GST_VAAPI_RATECONTROL_CQP)
+    return TRUE;
+
+  /* RateControl params */
+  misc = GST_VAAPI_ENC_MISC_PARAM_NEW (RateControl, encoder);
+  if (!misc)
+    return FALSE;
+  memcpy (misc->data, &GST_VAAPI_ENCODER_VA_RATE_CONTROL (encoder),
+      sizeof (VAEncMiscParameterRateControl));
+  gst_vaapi_enc_picture_add_misc_param (picture, misc);
+  gst_vaapi_codec_object_replace (&misc, NULL);
+
+  /* HRD params */
+  misc = GST_VAAPI_ENC_MISC_PARAM_NEW (HRD, encoder);
+  if (!misc)
+    return FALSE;
+  memcpy (misc->data, &GST_VAAPI_ENCODER_VA_HRD (encoder),
+      sizeof (VAEncMiscParameterHRD));
+  gst_vaapi_enc_picture_add_misc_param (picture, misc);
+  gst_vaapi_codec_object_replace (&misc, NULL);
+
+  /* FrameRate params */
+  if (GST_VAAPI_ENCODER_VA_FRAME_RATE (encoder).framerate == 0)
+    return TRUE;
+
+  misc = GST_VAAPI_ENC_MISC_PARAM_NEW (FrameRate, encoder);
+  if (!misc)
+    return FALSE;
+  memcpy (misc->data, &GST_VAAPI_ENCODER_VA_FRAME_RATE (encoder),
+      sizeof (VAEncMiscParameterFrameRate));
+  gst_vaapi_enc_picture_add_misc_param (picture, misc);
+  gst_vaapi_codec_object_replace (&misc, NULL);
+
+  return TRUE;
+}
+
 /**
  * gst_vaapi_encoder_ref:
  * @encoder: a #GstVaapiEncoder
@@ -556,6 +644,32 @@ get_packed_headers (GstVaapiEncoder * encoder)
   return encoder->packed_headers;
 }
 
+static gboolean
+get_roi_capability (GstVaapiEncoder * encoder, guint * num_roi_supported)
+{
+#if VA_CHECK_VERSION(0,39,1)
+  VAConfigAttribValEncROI *roi_config;
+  guint value;
+
+  if (!get_config_attribute (encoder, VAConfigAttribEncROI, &value))
+    return FALSE;
+
+  roi_config = (VAConfigAttribValEncROI *) & value;
+
+  if (roi_config->bits.num_roi_regions == 0 ||
+      VA_ROI_RC_QP_DELTA_SUPPORT (roi_config) == 0)
+    return FALSE;
+
+  GST_INFO ("Support for ROI - number of regions supported: %d",
+      roi_config->bits.num_roi_regions);
+
+  *num_roi_supported = roi_config->bits.num_roi_regions;
+  return TRUE;
+#else
+  return FALSE;
+#endif
+}
+
 static inline gboolean
 is_chroma_type_supported (GstVaapiEncoder * encoder)
 {
@@ -616,7 +730,8 @@ init_context_info (GstVaapiEncoder * encoder, GstVaapiContextInfo * cip,
   if (cdata->codec == GST_VAAPI_CODEC_JPEG) {
     cip->entrypoint = GST_VAAPI_ENTRYPOINT_PICTURE_ENCODE;
   } else {
-    if (cip->entrypoint != GST_VAAPI_ENTRYPOINT_SLICE_ENCODE_LP)
+    if (cip->entrypoint != GST_VAAPI_ENTRYPOINT_SLICE_ENCODE_LP &&
+        cip->entrypoint != GST_VAAPI_ENTRYPOINT_SLICE_ENCODE_FEI)
       cip->entrypoint = GST_VAAPI_ENTRYPOINT_SLICE_ENCODE;
   }
   cip->chroma_type = get_default_chroma_type (encoder, cip);
@@ -633,6 +748,7 @@ set_context_info (GstVaapiEncoder * encoder)
   GstVaapiConfigInfoEncoder *const config = &cip->config.encoder;
   const GstVideoFormat format =
       GST_VIDEO_INFO_FORMAT (GST_VAAPI_ENCODER_VIDEO_INFO (encoder));
+  guint fei_function = config->fei_function;
 
   init_context_info (encoder, cip, get_profile (encoder));
 
@@ -646,6 +762,10 @@ set_context_info (GstVaapiEncoder * encoder)
   memset (config, 0, sizeof (*config));
   config->rc_mode = GST_VAAPI_ENCODER_RATE_CONTROL (encoder);
   config->packed_headers = get_packed_headers (encoder);
+  config->roi_capability =
+      get_roi_capability (encoder, &config->roi_num_supported);
+  config->fei_function = fei_function;
+
   return TRUE;
 
   /* ERRORS */
@@ -686,11 +806,35 @@ gst_vaapi_encoder_reconfigure_internal (GstVaapiEncoder * encoder)
   GstVideoInfo *const vip = GST_VAAPI_ENCODER_VIDEO_INFO (encoder);
   GstVaapiEncoderStatus status;
   GstVaapiVideoPool *pool;
-  guint codedbuf_size;
+  guint codedbuf_size, target_percentage;
+  guint fps_d, fps_n;
+#if VA_CHECK_VERSION(0,36,0)
+  guint quality_level_max = 0;
+#endif
+
+  fps_d = GST_VIDEO_INFO_FPS_D (vip);
+  fps_n = GST_VIDEO_INFO_FPS_N (vip);
 
   /* Generate a keyframe every second */
   if (!encoder->keyframe_period)
-    encoder->keyframe_period = (vip->fps_n + vip->fps_d - 1) / vip->fps_d;
+    encoder->keyframe_period = (fps_n + fps_d - 1) / fps_d;
+
+  /* Default frame rate parameter */
+  if (fps_d > 0 && fps_n > 0)
+    GST_VAAPI_ENCODER_VA_FRAME_RATE (encoder).framerate = fps_d << 16 | fps_n;
+
+  target_percentage =
+      (GST_VAAPI_ENCODER_RATE_CONTROL (encoder) == GST_VAAPI_RATECONTROL_CBR) ?
+      100 : 70;
+
+  /* *INDENT-OFF* */
+  /* Default values for rate control parameter */
+  GST_VAAPI_ENCODER_VA_RATE_CONTROL (encoder) = (VAEncMiscParameterRateControl) {
+    .bits_per_second = encoder->bitrate * 1000,
+    .target_percentage = target_percentage,
+    .window_size = 500,
+  };
+  /* *INDENT-ON* */
 
   status = klass->reconfigure (encoder);
   if (status != GST_VAAPI_ENCODER_STATUS_SUCCESS)
@@ -699,6 +843,25 @@ gst_vaapi_encoder_reconfigure_internal (GstVaapiEncoder * encoder)
   if (!gst_vaapi_encoder_ensure_context (encoder))
     goto error_reset_context;
 
+  /* Currently only FEI entrypoint needed this.
+   * FEI ENC+PAK requires two contexts where the first one is for ENC
+   * and the second one is for PAK */
+  if (klass->ensure_secondary_context
+      && !klass->ensure_secondary_context (encoder))
+    goto error_reset_secondary_context;
+
+#if VA_CHECK_VERSION(0,36,0)
+  if (get_config_attribute (encoder, VAConfigAttribEncQualityRange,
+          &quality_level_max) && quality_level_max > 0) {
+    GST_VAAPI_ENCODER_QUALITY_LEVEL (encoder) =
+        CLAMP (GST_VAAPI_ENCODER_QUALITY_LEVEL (encoder), 1, quality_level_max);
+  } else {
+    GST_VAAPI_ENCODER_QUALITY_LEVEL (encoder) = 0;
+  }
+  GST_INFO ("Quality level is fixed to %d",
+      GST_VAAPI_ENCODER_QUALITY_LEVEL (encoder));
+#endif
+
   codedbuf_size = encoder->codedbuf_pool ?
       gst_vaapi_coded_buffer_pool_get_buffer_size (GST_VAAPI_CODED_BUFFER_POOL
       (encoder)) : 0;
@@ -723,6 +886,11 @@ error_reset_context:
     GST_ERROR ("failed to update VA context");
     return GST_VAAPI_ENCODER_STATUS_ERROR_OPERATION_FAILED;
   }
+error_reset_secondary_context:
+  {
+    GST_ERROR ("failed to create/update secondary VA context");
+    return GST_VAAPI_ENCODER_STATUS_ERROR_OPERATION_FAILED;
+  }
 }
 
 /**
@@ -810,6 +978,10 @@ set_property (GstVaapiEncoder * encoder, gint prop_id, const GValue * value)
     case GST_VAAPI_ENCODER_PROP_TUNE:
       status = gst_vaapi_encoder_set_tuning (encoder, g_value_get_enum (value));
       break;
+    case GST_VAAPI_ENCODER_PROP_QUALITY_LEVEL:
+      status = gst_vaapi_encoder_set_quality_level (encoder,
+          g_value_get_uint (value));
+      break;
   }
   return status;
 
@@ -950,18 +1122,14 @@ gst_vaapi_encoder_set_bitrate (GstVaapiEncoder * encoder, guint bitrate)
 {
   g_return_val_if_fail (encoder != NULL, 0);
 
-  if (encoder->bitrate != bitrate && encoder->num_codedbuf_queued > 0)
-    goto error_operation_failed;
+  if (encoder->bitrate != bitrate && encoder->num_codedbuf_queued > 0) {
+    GST_INFO ("Bitrate is changed to %d on runtime", bitrate);
+    encoder->bitrate = bitrate;
+    return gst_vaapi_encoder_reconfigure_internal (encoder);
+  }
 
   encoder->bitrate = bitrate;
   return GST_VAAPI_ENCODER_STATUS_SUCCESS;
-
-  /* ERRORS */
-error_operation_failed:
-  {
-    GST_ERROR ("could not change bitrate value after encoding started");
-    return GST_VAAPI_ENCODER_STATUS_ERROR_OPERATION_FAILED;
-  }
 }
 
 /**
@@ -1035,6 +1203,42 @@ error_operation_failed:
   }
 }
 
+/**
+ * gst_vaapi_encoder_set_quality_level:
+ * @encoder: a #GstVaapiEncoder
+ * @quality_level: the encoder quality level
+ *
+ * Notifies the @encoder to use the supplied @quality_level value.
+ *
+ * Note: currently, the quality_level can only be specified before
+ * the last call to gst_vaapi_encoder_set_codec_state(), which shall
+ * occur before the first frame is encoded. Afterwards, any change to
+ * this parameter causes gst_vaapi_encoder_set_quality_level() to
+ * return @GST_VAAPI_ENCODER_STATUS_ERROR_OPERATION_FAILED.
+ *
+ * Return value: a #GstVaapiEncoderStatus
+ */
+GstVaapiEncoderStatus
+gst_vaapi_encoder_set_quality_level (GstVaapiEncoder * encoder,
+    guint quality_level)
+{
+  g_return_val_if_fail (encoder != NULL, 0);
+
+  if (GST_VAAPI_ENCODER_QUALITY_LEVEL (encoder) != quality_level
+      && encoder->num_codedbuf_queued > 0)
+    goto error_operation_failed;
+
+  GST_VAAPI_ENCODER_QUALITY_LEVEL (encoder) = quality_level;
+  return GST_VAAPI_ENCODER_STATUS_SUCCESS;
+
+  /* ERRORS */
+error_operation_failed:
+  {
+    GST_ERROR ("could not change quality level after encoding started");
+    return GST_VAAPI_ENCODER_STATUS_ERROR_OPERATION_FAILED;
+  }
+}
+
 /* Initialize default values for configurable properties */
 static gboolean
 gst_vaapi_encoder_init_properties (GstVaapiEncoder * encoder)
@@ -1119,6 +1323,9 @@ gst_vaapi_encoder_finalize (GstVaapiEncoder * encoder)
 
   klass->finalize (encoder);
 
+  if (encoder->roi_regions)
+    g_list_free_full (encoder->roi_regions, g_free);
+
   gst_vaapi_object_replace (&encoder->context, NULL);
   gst_vaapi_display_replace (&encoder->display, NULL);
   encoder->va_display = NULL;
@@ -1267,6 +1474,203 @@ gst_vaapi_encoder_get_surface_formats (GstVaapiEncoder * encoder,
   return formats;
 }
 
+/**
+ * gst_vaapi_encoder_ensure_num_slices:
+ * @encoder: a #GstVaapiEncoder
+ * @profile: a #GstVaapiProfile
+ * @entrypoint: a #GstVaapiEntrypoint
+ * @media_max_slices: the number of the slices permitted by the stream
+ * @num_slices: (out): the possible number of slices to process
+ *
+ * This function will clamp the @num_slices provided by the user,
+ * according the limit of the number of slices permitted by the stream
+ * and by the hardware.
+ *
+ * We need to pass the @profile and the @entrypoint, because at the
+ * moment the encoder base class, still doesn't have them assigned,
+ * and this function is meant to be called by the derived classes
+ * while they are configured.
+ *
+ * Returns: %TRUE if the number of slices is different than zero.
+ **/
+gboolean
+gst_vaapi_encoder_ensure_num_slices (GstVaapiEncoder * encoder,
+    GstVaapiProfile profile, GstVaapiEntrypoint entrypoint,
+    guint media_max_slices, guint * num_slices)
+{
+  VAProfile va_profile;
+  VAEntrypoint va_entrypoint;
+  guint max_slices, num;
+
+  va_profile = gst_vaapi_profile_get_va_profile (profile);
+  va_entrypoint = gst_vaapi_entrypoint_get_va_entrypoint (entrypoint);
+
+  if (!gst_vaapi_get_config_attribute (encoder->display, va_profile,
+          va_entrypoint, VAConfigAttribEncMaxSlices, &max_slices)) {
+    *num_slices = 1;
+    return TRUE;
+  }
+
+  num = *num_slices;
+  if (num > max_slices)
+    num = max_slices;
+  if (num > media_max_slices)
+    num = media_max_slices;
+
+  if (num == 0)
+    return FALSE;
+  *num_slices = num;
+  return TRUE;
+}
+
+/**
+ * gst_vaapi_encoder_ensure_max_num_ref_frames:
+ * @encoder: a #GstVaapiEncoder
+ * @profile: a #GstVaapiProfile
+ * @entrypoint: a #GstVaapiEntrypoint
+ *
+ * This function will query VAConfigAttribEncMaxRefFrames to get the
+ * maximum number of reference frames in the driver,
+ * for both the reference picture list 0 (bottom 16 bits) and
+ * the reference picture list 1 (top 16 bits).
+ *
+ * We need to pass the @profile and the @entrypoint, because at the
+ * moment the encoder base class, still doesn't have them assigned,
+ * and this function is meant to be called by the derived classes
+ * while they are configured.
+ *
+ * Returns: %TRUE if the number of reference frames is different than zero.
+ **/
+gboolean
+gst_vaapi_encoder_ensure_max_num_ref_frames (GstVaapiEncoder * encoder,
+    GstVaapiProfile profile, GstVaapiEntrypoint entrypoint)
+{
+  VAProfile va_profile;
+  VAEntrypoint va_entrypoint;
+  guint max_ref_frames;
+
+  va_profile = gst_vaapi_profile_get_va_profile (profile);
+  va_entrypoint = gst_vaapi_entrypoint_get_va_entrypoint (entrypoint);
+
+  if (!gst_vaapi_get_config_attribute (encoder->display, va_profile,
+          va_entrypoint, VAConfigAttribEncMaxRefFrames, &max_ref_frames)) {
+    /* Set the default the number of reference frames */
+    encoder->max_num_ref_frames_0 = 1;
+    encoder->max_num_ref_frames_1 = 0;
+    return TRUE;
+  }
+
+  encoder->max_num_ref_frames_0 = max_ref_frames & 0xffff;
+  encoder->max_num_ref_frames_1 = (max_ref_frames >> 16) & 0xffff;
+
+  return TRUE;
+}
+
+/**
+ * gst_vaapi_encoder_add_roi:
+ * @encoder: a #GstVaapiEncoder
+ * @roi: (transfer none): a #GstVaapiROI
+ *
+ * Adds a roi region provided by user.
+ *
+ * This can be called on running a pipeline,
+ * Since vaapi encoder set roi regions at every frame encoding.
+ * Note that if it exceeds number of supported roi in driver,
+ * this will return FALSE.
+ *
+ * Return value: a #gboolean
+ */
+gboolean
+gst_vaapi_encoder_add_roi (GstVaapiEncoder * encoder, GstVaapiROI * roi)
+{
+  GstVaapiContextInfo *const cip = &encoder->context_info;
+  const GstVaapiConfigInfoEncoder *const config = &cip->config.encoder;
+  GstVaapiROI *region = NULL;
+  GList *walk;
+
+  g_return_val_if_fail (roi != NULL, FALSE);
+
+  if (!config->roi_capability)
+    return FALSE;
+
+  if (encoder->roi_regions &&
+      g_list_length (encoder->roi_regions) > config->roi_num_supported)
+    return FALSE;
+
+  walk = encoder->roi_regions;
+  while (walk) {
+    GstVaapiROI *region_ptr = (GstVaapiROI *) walk->data;
+    if (region_ptr->rect.x == roi->rect.x &&
+        region_ptr->rect.y == roi->rect.y &&
+        region_ptr->rect.width == roi->rect.width &&
+        region_ptr->rect.height == roi->rect.height) {
+      /* Duplicated region */
+      goto end;
+    }
+    walk = walk->next;
+  }
+
+  region = g_malloc0 (sizeof (GstVaapiROI));
+  if (G_UNLIKELY (!region))
+    return FALSE;
+
+  region->rect.x = roi->rect.x;
+  region->rect.y = roi->rect.y;
+  region->rect.width = roi->rect.width;
+  region->rect.height = roi->rect.height;
+
+  encoder->roi_regions = g_list_append (encoder->roi_regions, region);
+
+end:
+  return TRUE;
+}
+
+/**
+ * gst_vaapi_encoder_del_roi:
+ * @encoder: a #GstVaapiEncoder
+ * @roi: (transfer none): a #GstVaapiROI
+ *
+ * Deletes a roi region provided by user.
+ *
+ * This can be called on running a pipeline,
+ * Since vaapi encoder set roi regions at every frame encoding.
+ *
+ * Return value: a #gboolean
+ */
+gboolean
+gst_vaapi_encoder_del_roi (GstVaapiEncoder * encoder, GstVaapiROI * roi)
+{
+  GstVaapiContextInfo *const cip = &encoder->context_info;
+  const GstVaapiConfigInfoEncoder *const config = &cip->config.encoder;
+  GList *walk;
+  gboolean ret = FALSE;
+
+  g_return_val_if_fail (roi != NULL, FALSE);
+
+  if (!config->roi_capability)
+    return FALSE;
+
+  if (encoder->roi_regions && g_list_length (encoder->roi_regions) == 0)
+    return FALSE;
+
+  walk = encoder->roi_regions;
+  while (walk) {
+    GstVaapiROI *region = (GstVaapiROI *) walk->data;
+    if (region->rect.x == roi->rect.x &&
+        region->rect.y == roi->rect.y &&
+        region->rect.width == roi->rect.width &&
+        region->rect.height == roi->rect.height) {
+      encoder->roi_regions = g_list_remove (encoder->roi_regions, region);
+      g_free (region);
+      ret = TRUE;
+      break;
+    }
+    walk = walk->next;
+  }
+
+  return ret;
+}
+
 /** Returns a GType for the #GstVaapiEncoderTune set */
 GType
 gst_vaapi_encoder_tune_get_type (void)
@@ -1294,3 +1698,25 @@ gst_vaapi_encoder_tune_get_type (void)
   }
   return g_type;
 }
+
+/** Returns a GType for the #GstVaapiEncoderMbbrc set */
+GType
+gst_vaapi_encoder_mbbrc_get_type (void)
+{
+  static volatile gsize g_type = 0;
+
+  if (g_once_init_enter (&g_type)) {
+    static const GEnumValue encoder_mbbrc_values[] = {
+      {GST_VAAPI_ENCODER_MBBRC_AUTO, "Auto", "auto"},
+      {GST_VAAPI_ENCODER_MBBRC_ON, "On", "on"},
+      {GST_VAAPI_ENCODER_MBBRC_OFF, "Off", "off"},
+      {0, NULL, NULL},
+    };
+
+    GType type =
+        g_enum_register_static (g_intern_static_string ("GstVaapiEncoderMbbrc"),
+        encoder_mbbrc_values);
+    g_once_init_leave (&g_type, type);
+  }
+  return g_type;
+}
