@@ -93,6 +93,23 @@ typedef enum {
   GST_VAAPI_ENCODER_TUNE_LOW_POWER,
 } GstVaapiEncoderTune;
 
+/**
+ * GstVaapiEncoderMbbrc:
+ * @GST_VAAPI_ENCODER_MBBRC_AUTO: bitrate control auto
+ * @GST_VAAPI_ENCODER_MBBRC_ON: bitrate control on
+ * @GST_VAAPI_ENCODER_MBBRC_OFF: bitrate control off
+ *
+ * Values for the macroblock level bitrate control.
+ *
+ * This property values are only available for H264 and H265 (HEVC)
+ * encoders, when rate control is not Constant QP.
+ **/
+typedef enum {
+  GST_VAAPI_ENCODER_MBBRC_AUTO = 0,
+  GST_VAAPI_ENCODER_MBBRC_ON = 1,
+  GST_VAAPI_ENCODER_MBBRC_OFF = 2,
+} GstVaapiEncoderMbbrc;
+
 /**
  * GstVaapiEncoderProp:
  * @GST_VAAPI_ENCODER_PROP_RATECONTROL: Rate control (#GstVaapiRateControl).
@@ -108,6 +125,7 @@ typedef enum {
   GST_VAAPI_ENCODER_PROP_BITRATE,
   GST_VAAPI_ENCODER_PROP_KEYFRAME_PERIOD,
   GST_VAAPI_ENCODER_PROP_TUNE,
+  GST_VAAPI_ENCODER_PROP_QUALITY_LEVEL
 } GstVaapiEncoderProp;
 
 /**
@@ -122,9 +140,17 @@ typedef struct {
   GParamSpec *const pspec;
 } GstVaapiEncoderPropInfo;
 
+typedef struct _GstVaapiROI {
+  gint roi_value;
+  GstVaapiRectangle rect;
+} GstVaapiROI;
+
 GType
 gst_vaapi_encoder_tune_get_type (void) G_GNUC_CONST;
 
+GType
+gst_vaapi_encoder_mbbrc_get_type (void) G_GNUC_CONST;
+
 GstVaapiEncoder *
 gst_vaapi_encoder_ref (GstVaapiEncoder * encoder);
 
@@ -166,6 +192,10 @@ GstVaapiEncoderStatus
 gst_vaapi_encoder_set_tuning (GstVaapiEncoder * encoder,
     GstVaapiEncoderTune tuning);
 
+GstVaapiEncoderStatus
+gst_vaapi_encoder_set_quality_level (GstVaapiEncoder * encoder,
+    guint quality_level);
+
 GstVaapiEncoderStatus
 gst_vaapi_encoder_get_buffer_with_timeout (GstVaapiEncoder * encoder,
     GstVaapiCodedBufferProxy ** out_codedbuf_proxy_ptr, guint64 timeout);
@@ -176,6 +206,13 @@ gst_vaapi_encoder_flush (GstVaapiEncoder * encoder);
 GArray *
 gst_vaapi_encoder_get_surface_formats (GstVaapiEncoder * encoder,
     GstVaapiProfile profile);
+
+gboolean
+gst_vaapi_encoder_add_roi (GstVaapiEncoder * encoder, GstVaapiROI * roi);
+
+gboolean
+gst_vaapi_encoder_del_roi (GstVaapiEncoder * encoder, GstVaapiROI * roi);
+
 G_END_DECLS
 
 #endif /* GST_VAAPI_ENCODER_H */
