@@ -22,8 +22,6 @@
 
 #include "sysdeps.h"
 #include <math.h>
-#include <va/va.h>
-#include <va/va_enc_hevc.h>
 #include <gst/base/gstbitwriter.h>
 #include <gst/codecparsers/gsth265parser.h>
 #include "gstvaapicompat.h"
@@ -40,8 +38,9 @@
 
 /* Supported set of VA rate controls, within this implementation */
 #define SUPPORTED_RATECONTROLS                          \
-  (GST_VAAPI_RATECONTROL_MASK (CQP)) |                  \
-  GST_VAAPI_RATECONTROL_MASK (CBR)
+  (GST_VAAPI_RATECONTROL_MASK (CQP) |                   \
+   GST_VAAPI_RATECONTROL_MASK (CBR) |                   \
+   GST_VAAPI_RATECONTROL_MASK (VBR))
 
 /* Supported set of tuning options, within this implementation */
 #define SUPPORTED_TUNE_OPTIONS                          \
@@ -86,9 +85,6 @@ typedef struct _GstVaapiH265ReorderPool
 /* --- H.265 Encoder                                                     --- */
 /* ------------------------------------------------------------------------- */
 
-#define GST_VAAPI_ENCODER_H265_CAST(encoder) \
-    ((GstVaapiEncoderH265 *)(encoder))
-
 struct _GstVaapiEncoderH265
 {
   GstVaapiEncoder parent_instance;
@@ -103,6 +99,9 @@ struct _GstVaapiEncoderH265
   guint32 idr_period;
   guint32 init_qp;
   guint32 min_qp;
+  guint32 qp_i;
+  guint32 qp_ip;
+  guint32 qp_ib;
   guint32 num_slices;
   guint32 num_bframes;
   guint32 ctu_width;            /* CTU == Coding Tree Unit */
@@ -122,6 +121,7 @@ struct _GstVaapiEncoderH265
   guint32 max_pic_order_cnt;
   guint32 log2_max_pic_order_cnt;
   guint32 idr_num;
+  guint num_ref_frames;
 
   GstBuffer *vps_data;
   GstBuffer *sps_data;
@@ -130,6 +130,7 @@ struct _GstVaapiEncoderH265
   guint bitrate_bits;           // bitrate (bits)
   guint cpb_length;             // length of CPB buffer (ms)
   guint cpb_length_bits;        // length of CPB buffer (bits)
+  GstVaapiEncoderMbbrc mbbrc;   // macroblock bitrate control
 
   /* Crop rectangle */
   guint conformance_window_flag:1;
@@ -665,13 +666,14 @@ bs_write_slice (GstBitWriter * bs,
   guint8 no_output_of_prior_pics_flag = 0;
   guint8 dependent_slice_segment_flag = 0;
   guint8 short_term_ref_pic_set_sps_flag = 0;
-  guint8 num_ref_idx_active_override_flag = 0;
   guint8 slice_deblocking_filter_disabled_flag = 0;
+  guint8 num_ref_idx_active_override_flag =
+      slice_param->slice_fields.bits.num_ref_idx_active_override_flag;
 
   /* first_slice_segment_in_pic_flag */
   WRITE_UINT32 (bs, encoder->first_slice_segment_in_pic_flag, 1);
 
-  /* Fixme: For all IRAP pics */
+  /* FIXME: For all IRAP pics */
   /* no_output_of_prior_pics_flag */
   if (GST_VAAPI_ENC_PICTURE_IS_IDR (picture))
     WRITE_UINT32 (bs, no_output_of_prior_pics_flag, 1);
@@ -701,10 +703,29 @@ bs_write_slice (GstBitWriter * bs,
         guint num_positive_pics = 0, num_negative_pics = 0;
         guint delta_poc_s0_minus1 = 0, delta_poc_s1_minus1 = 0;
         guint used_by_curr_pic_s0_flag = 0, used_by_curr_pic_s1_flag = 0;
+        guint reflist_0_count = 0, reflist_1_count = 0;
+        gint i;
+
+        /* Get count of ref_pic_list */
+        if (picture->type == GST_VAAPI_PICTURE_TYPE_P
+            || picture->type == GST_VAAPI_PICTURE_TYPE_B) {
+          for (i = 0; i < G_N_ELEMENTS (slice_param->ref_pic_list0); ++i) {
+            if (slice_param->ref_pic_list0[i].picture_id == VA_INVALID_SURFACE)
+              break;
+          }
+          reflist_0_count = i;
+
+          if (picture->type == GST_VAAPI_PICTURE_TYPE_B) {
+            for (i = 0; i < G_N_ELEMENTS (slice_param->ref_pic_list1); ++i) {
+              if (slice_param->ref_pic_list1[i].picture_id ==
+                  VA_INVALID_SURFACE)
+                break;
+            }
+            reflist_1_count = i;
+          }
+        }
 
         if (picture->type == GST_VAAPI_PICTURE_TYPE_P) {
-          num_negative_pics = 1;
-          num_positive_pics = 0;
           delta_poc_s0_minus1 =
               picture->poc - slice_param->ref_pic_list0[0].pic_order_cnt - 1;
           used_by_curr_pic_s0_flag = 1;
@@ -712,8 +733,6 @@ bs_write_slice (GstBitWriter * bs,
           used_by_curr_pic_s1_flag = 0;
         }
         if (picture->type == GST_VAAPI_PICTURE_TYPE_B) {
-          num_negative_pics = 1;
-          num_positive_pics = 1;
           delta_poc_s0_minus1 =
               picture->poc - slice_param->ref_pic_list0[0].pic_order_cnt - 1;
           used_by_curr_pic_s0_flag = 1;
@@ -722,19 +741,35 @@ bs_write_slice (GstBitWriter * bs,
           used_by_curr_pic_s1_flag = 1;
         }
 
+        num_negative_pics = reflist_0_count;
+        num_positive_pics = reflist_1_count;
+
         /* num_negative_pics */
         WRITE_UE (bs, num_negative_pics);
         /* num_positive_pics */
         WRITE_UE (bs, num_positive_pics);
-        if (num_negative_pics) {
+
+        for (i = 0; i < num_negative_pics; i++) {
           /* delta_poc_s0_minus1 */
-          WRITE_UE (bs, delta_poc_s0_minus1);
+          if (i == 0) {
+            WRITE_UE (bs, delta_poc_s0_minus1);
+          } else {
+            WRITE_UE (bs,
+                slice_param->ref_pic_list0[i - 1].pic_order_cnt -
+                slice_param->ref_pic_list0[i].pic_order_cnt - 1);
+          }
           /* used_by_curr_pic_s0_flag */
           WRITE_UINT32 (bs, used_by_curr_pic_s0_flag, 1);
         }
-        if (num_positive_pics) {
+        for (i = 0; i < num_positive_pics; i++) {
           /* delta_poc_s1_minus1 */
-          WRITE_UE (bs, delta_poc_s1_minus1);
+          if (i == 0) {
+            WRITE_UE (bs, delta_poc_s1_minus1);
+          } else {
+            WRITE_UE (bs,
+                slice_param->ref_pic_list1[i - 1].pic_order_cnt -
+                slice_param->ref_pic_list1[i].pic_order_cnt - 1);
+          }
           /* used_by_curr_pic_s1_flag */
           WRITE_UINT32 (bs, used_by_curr_pic_s1_flag, 1);
         }
@@ -756,6 +791,12 @@ bs_write_slice (GstBitWriter * bs,
         slice_param->slice_type == GST_H265_B_SLICE) {
       /* num_ref_idx_active_override_flag */
       WRITE_UINT32 (bs, num_ref_idx_active_override_flag, 1);
+      if (num_ref_idx_active_override_flag) {
+        WRITE_UE (bs, slice_param->num_ref_idx_l0_active_minus1);
+        if (slice_param->slice_type == GST_H265_B_SLICE)
+          WRITE_UE (bs, slice_param->num_ref_idx_l1_active_minus1);
+      }
+
       /* mvd_l1_zero_flag */
       if (slice_param->slice_type == GST_H265_B_SLICE)
         WRITE_UINT32 (bs, slice_param->slice_fields.bits.mvd_l1_zero_flag, 1);
@@ -906,7 +947,8 @@ ensure_hw_profile (GstVaapiEncoderH265 * encoder)
   /* ERRORS */
 error_unsupported_profile:
   {
-    GST_ERROR ("unsupported HW profile (0x%08x)", encoder->profile);
+    GST_ERROR ("unsupported HW profile %s",
+        gst_vaapi_profile_get_name (encoder->profile));
     return FALSE;
   }
 }
@@ -957,7 +999,8 @@ ensure_tier (GstVaapiEncoderH265 * encoder)
 {
 
   encoder->tier = GST_VAAPI_TIER_H265_MAIN;
-  /*Fixme: Derive proper tier based on upstream caps or limits, coding tools etc */
+  /* FIXME: Derive proper tier based on upstream caps or limits, coding
+   * tools etc */
 
   return TRUE;
 }
@@ -976,8 +1019,8 @@ ensure_level (GstVaapiEncoderH265 * encoder)
     const GstVaapiH265LevelLimits *const limits = &limits_table[i];
     if (PicSizeInSamplesY <= limits->MaxLumaPs)
       break;
-    /* Fixme: Add more constraint checking:tier (extracted from caps), cpb size,
-     * bitrate, num_tile_columns and num_tile_rows */
+    /* FIXME: Add more constraint checking:tier (extracted from caps),
+     * cpb size, bitrate, num_tile_columns and num_tile_rows */
   }
   if (i == num_limits)
     goto error_unsupported_level;
@@ -1448,7 +1491,9 @@ fill_sequence (GstVaapiEncoderH265 * encoder, GstVaapiEncSequence * sequence)
 
   seq_param->general_profile_idc = encoder->profile_idc;
   seq_param->general_level_idc = encoder->level_idc;
-  seq_param->general_tier_flag = 0;     /* Fixme: use the tier flag extracted from upstream caps or calcuted one */
+  seq_param->general_tier_flag = 0;     /* FIXME: use the tier flag
+                                         * extracted from upstream
+                                         * caps or calcuted one */
 
   seq_param->intra_period = GST_VAAPI_ENCODER_KEYFRAME_PERIOD (encoder);
   seq_param->intra_idr_period = encoder->idr_period;
@@ -1537,6 +1582,7 @@ fill_picture (GstVaapiEncoderH265 * encoder, GstVaapiEncPicture * picture,
 
       pic_param->reference_frames[i].picture_id =
           GST_VAAPI_SURFACE_PROXY_SURFACE_ID (ref_pic->pic);
+      pic_param->reference_frames[i].pic_order_cnt = ref_pic->poc;
       ++i;
     }
     g_assert (i <= 15 && i <= ref_pool->max_ref_frames);
@@ -1551,7 +1597,7 @@ fill_picture (GstVaapiEncoderH265 * encoder, GstVaapiEncPicture * picture,
   pic_param->collocated_ref_pic_index = 0xFF;
 
   pic_param->last_picture = 0;
-  pic_param->pic_init_qp = encoder->init_qp;
+  pic_param->pic_init_qp = encoder->qp_i;
   pic_param->num_ref_idx_l0_default_active_minus1 =
       (ref_pool->max_reflist0_count ? (ref_pool->max_reflist0_count - 1) : 0);
   pic_param->num_ref_idx_l1_default_active_minus1 =
@@ -1572,8 +1618,9 @@ fill_picture (GstVaapiEncoderH265 * encoder, GstVaapiEncPicture * picture,
   pic_param->pic_fields.bits.transform_skip_enabled_flag = TRUE;
   /* it seems driver requires enablement of cu_qp_delta_enabled_flag
    * to modifiy QP values in CBR mode encoding */
-  if (GST_VAAPI_ENCODER_RATE_CONTROL (encoder) == GST_VAAPI_RATECONTROL_CBR)
-    pic_param->pic_fields.bits.cu_qp_delta_enabled_flag = TRUE;
+  pic_param->pic_fields.bits.cu_qp_delta_enabled_flag =
+      GST_VAAPI_ENCODER_RATE_CONTROL (encoder) != GST_VAAPI_RATECONTROL_CQP;
+
   pic_param->pic_fields.bits.pps_loop_filter_across_slices_enabled_flag = TRUE;
 
   if (GST_VAAPI_ENC_PICTURE_IS_IDR (picture))
@@ -1639,6 +1686,8 @@ add_slice_headers (GstVaapiEncoderH265 * encoder, GstVaapiEncPicture * picture,
     slice_param->slice_type = h265_get_slice_type (picture->type);
     slice_param->slice_pic_parameter_set_id = 0;
 
+    slice_param->slice_fields.bits.num_ref_idx_active_override_flag =
+        reflist_0_count || reflist_1_count;
     if (picture->type != GST_VAAPI_PICTURE_TYPE_I && reflist_0_count > 0)
       slice_param->num_ref_idx_l0_active_minus1 = reflist_0_count - 1;
     else
@@ -1647,8 +1696,6 @@ add_slice_headers (GstVaapiEncoderH265 * encoder, GstVaapiEncPicture * picture,
       slice_param->num_ref_idx_l1_active_minus1 = reflist_1_count - 1;
     else
       slice_param->num_ref_idx_l1_active_minus1 = 0;
-    g_assert (slice_param->num_ref_idx_l0_active_minus1 == 0);
-    g_assert (slice_param->num_ref_idx_l1_active_minus1 == 0);
 
     i_ref = 0;
     if (picture->type != GST_VAAPI_PICTURE_TYPE_I) {
@@ -1657,7 +1704,6 @@ add_slice_headers (GstVaapiEncoderH265 * encoder, GstVaapiEncPicture * picture,
             GST_VAAPI_SURFACE_PROXY_SURFACE_ID (reflist_0[i_ref]->pic);
         slice_param->ref_pic_list0[i_ref].pic_order_cnt = reflist_0[i_ref]->poc;
       }
-      g_assert (i_ref == 1);
     }
     for (; i_ref < G_N_ELEMENTS (slice_param->ref_pic_list0); ++i_ref) {
       slice_param->ref_pic_list0[i_ref].picture_id = VA_INVALID_SURFACE;
@@ -1671,7 +1717,6 @@ add_slice_headers (GstVaapiEncoderH265 * encoder, GstVaapiEncPicture * picture,
             GST_VAAPI_SURFACE_PROXY_SURFACE_ID (reflist_1[i_ref]->pic);
         slice_param->ref_pic_list1[i_ref].pic_order_cnt = reflist_1[i_ref]->poc;
       }
-      g_assert (i_ref == 1);
     }
     for (; i_ref < G_N_ELEMENTS (slice_param->ref_pic_list1); ++i_ref) {
       slice_param->ref_pic_list1[i_ref].picture_id = VA_INVALID_SURFACE;
@@ -1679,9 +1724,22 @@ add_slice_headers (GstVaapiEncoderH265 * encoder, GstVaapiEncPicture * picture,
     }
 
     slice_param->max_num_merge_cand = 5;        /* MaxNumMergeCand  */
-    slice_param->slice_qp_delta = encoder->init_qp - encoder->min_qp;
-
-    slice_param->slice_fields.value = 0;
+    slice_param->slice_qp_delta = encoder->qp_i - encoder->init_qp;
+    if (GST_VAAPI_ENCODER_RATE_CONTROL (encoder) == GST_VAAPI_RATECONTROL_CQP) {
+      if (picture->type == GST_VAAPI_PICTURE_TYPE_P) {
+        slice_param->slice_qp_delta += encoder->qp_ip;
+      } else if (picture->type == GST_VAAPI_PICTURE_TYPE_B) {
+        slice_param->slice_qp_delta += encoder->qp_ib;
+      }
+      if ((gint) encoder->init_qp + slice_param->slice_qp_delta <
+          (gint) encoder->min_qp) {
+        slice_param->slice_qp_delta = encoder->min_qp - encoder->init_qp;
+      }
+      /* TODO: max_qp could be provided as a property in the future */
+      if ((gint) encoder->init_qp + slice_param->slice_qp_delta > 51) {
+        slice_param->slice_qp_delta = 51 - encoder->init_qp;
+      }
+    }
 
     slice_param->slice_fields.bits.
         slice_loop_filter_across_slices_enabled_flag = TRUE;
@@ -1761,22 +1819,41 @@ error_create_packed_seq_hdr:
   }
 }
 
+static gboolean
+ensure_control_rate_params (GstVaapiEncoderH265 * encoder)
+{
+  if (GST_VAAPI_ENCODER_RATE_CONTROL (encoder) == GST_VAAPI_RATECONTROL_CQP)
+    return TRUE;
+
+  /* RateControl params */
+  GST_VAAPI_ENCODER_VA_RATE_CONTROL (encoder).bits_per_second =
+      encoder->bitrate_bits;
+  /* CPB (Coded picture buffer) length in milliseconds, which could be
+   * provided as a property */
+  GST_VAAPI_ENCODER_VA_RATE_CONTROL (encoder).window_size = encoder->cpb_length;
+  GST_VAAPI_ENCODER_VA_RATE_CONTROL (encoder).initial_qp = encoder->init_qp;
+  GST_VAAPI_ENCODER_VA_RATE_CONTROL (encoder).min_qp = encoder->min_qp;
+
+#if VA_CHECK_VERSION(1,0,0)
+  GST_VAAPI_ENCODER_VA_RATE_CONTROL (encoder).rc_flags.bits.mb_rate_control =
+      (guint) encoder->mbbrc;
+#endif
+
+  /* HRD params */
+  fill_hrd_params (encoder, &GST_VAAPI_ENCODER_VA_HRD (encoder));
+
+  return TRUE;
+}
+
 static gboolean
 ensure_misc_params (GstVaapiEncoderH265 * encoder, GstVaapiEncPicture * picture)
 {
-  GstVaapiEncMiscParam *misc = NULL;
-
-  /* HRD params for rate control */
-  if (GST_VAAPI_ENCODER_RATE_CONTROL (encoder) == GST_VAAPI_RATECONTROL_CBR) {
-    misc = GST_VAAPI_ENC_MISC_PARAM_NEW (HRD, encoder);
-    g_assert (misc);
-    if (!misc)
-      return FALSE;
-    fill_hrd_params (encoder, misc->data);
-    gst_vaapi_enc_picture_add_misc_param (picture, misc);
-    gst_vaapi_codec_object_replace (&misc, NULL);
-  }
+  GstVaapiEncoder *const base_encoder = GST_VAAPI_ENCODER_CAST (encoder);
 
+  if (!gst_vaapi_encoder_ensure_param_control_rate (base_encoder, picture))
+    return FALSE;
+  if (!gst_vaapi_encoder_ensure_param_quality_level (base_encoder, picture))
+    return FALSE;
   return TRUE;
 }
 
@@ -1876,8 +1953,9 @@ ensure_bitrate (GstVaapiEncoderH265 * encoder)
 
   switch (GST_VAAPI_ENCODER_RATE_CONTROL (encoder)) {
     case GST_VAAPI_RATECONTROL_CBR:
+    case GST_VAAPI_RATECONTROL_VBR:
       if (!base_encoder->bitrate) {
-        /* Fixme: Provide better estimation */
+        /* FIXME: Provide better estimation */
         /* Using a 1/6 compression ratio */
         /* 12 bits per pixel for YUV420 */
         guint64 factor;
@@ -1949,18 +2027,28 @@ reset_properties (GstVaapiEncoderH265 * encoder)
   if (encoder->idr_period > MAX_IDR_PERIOD)
     encoder->idr_period = MAX_IDR_PERIOD;
 
-  /*Fixme: provide user control for idr_period ?? */
-  encoder->idr_period = base_encoder->keyframe_period * 2;
-
-  if (encoder->min_qp > encoder->init_qp ||
-      (GST_VAAPI_ENCODER_RATE_CONTROL (encoder) == GST_VAAPI_RATECONTROL_CQP &&
-          encoder->min_qp < encoder->init_qp))
+  if (encoder->min_qp > encoder->init_qp)
     encoder->min_qp = encoder->init_qp;
+  encoder->qp_i = encoder->init_qp;
 
   ctu_size = encoder->ctu_width * encoder->ctu_height;
-  if (encoder->num_slices > (ctu_size + 1) / 2)
-    encoder->num_slices = (ctu_size + 1) / 2;
-  g_assert (encoder->num_slices);
+  g_assert (gst_vaapi_encoder_ensure_num_slices (base_encoder, encoder->profile,
+          GST_VAAPI_ENTRYPOINT_SLICE_ENCODE, (ctu_size + 1) / 2,
+          &encoder->num_slices));
+
+  gst_vaapi_encoder_ensure_max_num_ref_frames (base_encoder, encoder->profile,
+      GST_VAAPI_ENTRYPOINT_SLICE_ENCODE);
+
+  if (base_encoder->max_num_ref_frames_1 < 1 && encoder->num_bframes > 0) {
+    GST_WARNING ("Disabling b-frame since the driver doesn't support it");
+    encoder->num_bframes = 0;
+  }
+
+  if (encoder->num_ref_frames > base_encoder->max_num_ref_frames_0) {
+    GST_INFO ("Lowering the number of reference frames to %d",
+        base_encoder->max_num_ref_frames_0);
+    encoder->num_ref_frames = base_encoder->max_num_ref_frames_0;
+  }
 
   if (encoder->num_bframes > (base_encoder->keyframe_period + 1) / 2)
     encoder->num_bframes = (base_encoder->keyframe_period + 1) / 2;
@@ -1980,16 +2068,15 @@ reset_properties (GstVaapiEncoderH265 * encoder)
 
   /* Only Supporting a maximum of two reference frames */
   if (encoder->num_bframes) {
-    encoder->max_dec_pic_buffering = 3;
+    encoder->max_dec_pic_buffering = encoder->num_ref_frames + 2;
     encoder->max_num_reorder_pics = 1;
   } else {
-    encoder->max_dec_pic_buffering =
-        (GST_VAAPI_ENCODER_KEYFRAME_PERIOD (encoder) == 1) ? 1 : 2;
+    encoder->max_dec_pic_buffering = encoder->num_ref_frames + 1;
     encoder->max_num_reorder_pics = 0;
   }
 
   ref_pool = &encoder->ref_pool;
-  ref_pool->max_reflist0_count = 1;
+  ref_pool->max_reflist0_count = encoder->num_ref_frames;
   ref_pool->max_reflist1_count = encoder->num_bframes > 0;
   ref_pool->max_ref_frames = ref_pool->max_reflist0_count
       + ref_pool->max_reflist1_count;
@@ -2002,8 +2089,7 @@ static GstVaapiEncoderStatus
 gst_vaapi_encoder_h265_encode (GstVaapiEncoder * base_encoder,
     GstVaapiEncPicture * picture, GstVaapiCodedBufferProxy * codedbuf)
 {
-  GstVaapiEncoderH265 *const encoder =
-      GST_VAAPI_ENCODER_H265_CAST (base_encoder);
+  GstVaapiEncoderH265 *const encoder = GST_VAAPI_ENCODER_H265 (base_encoder);
   GstVaapiEncoderStatus ret = GST_VAAPI_ENCODER_STATUS_ERROR_UNKNOWN;
   GstVaapiSurfaceProxy *reconstruct = NULL;
 
@@ -2040,8 +2126,7 @@ error:
 static GstVaapiEncoderStatus
 gst_vaapi_encoder_h265_flush (GstVaapiEncoder * base_encoder)
 {
-  GstVaapiEncoderH265 *const encoder =
-      GST_VAAPI_ENCODER_H265_CAST (base_encoder);
+  GstVaapiEncoderH265 *const encoder = GST_VAAPI_ENCODER_H265 (base_encoder);
   GstVaapiH265ReorderPool *reorder_pool;
   GstVaapiEncPicture *pic;
 
@@ -2064,8 +2149,7 @@ static GstVaapiEncoderStatus
 gst_vaapi_encoder_h265_get_codec_data (GstVaapiEncoder * base_encoder,
     GstBuffer ** out_buffer_ptr)
 {
-  GstVaapiEncoderH265 *const encoder =
-      GST_VAAPI_ENCODER_H265_CAST (base_encoder);
+  GstVaapiEncoderH265 *const encoder = GST_VAAPI_ENCODER_H265 (base_encoder);
   const guint32 configuration_version = 0x01;
   const guint32 nal_length_size = 4;
   GstMapInfo vps_info, sps_info, pps_info;
@@ -2211,8 +2295,7 @@ static GstVaapiEncoderStatus
 gst_vaapi_encoder_h265_reordering (GstVaapiEncoder * base_encoder,
     GstVideoCodecFrame * frame, GstVaapiEncPicture ** output)
 {
-  GstVaapiEncoderH265 *const encoder =
-      GST_VAAPI_ENCODER_H265_CAST (base_encoder);
+  GstVaapiEncoderH265 *const encoder = GST_VAAPI_ENCODER_H265 (base_encoder);
   GstVaapiH265ReorderPool *reorder_pool = NULL;
   GstVaapiEncPicture *picture;
   gboolean is_idr = FALSE;
@@ -2311,13 +2394,13 @@ end:
 static GstVaapiEncoderStatus
 set_context_info (GstVaapiEncoder * base_encoder)
 {
-  GstVaapiEncoderH265 *const encoder =
-      GST_VAAPI_ENCODER_H265_CAST (base_encoder);
+  GstVaapiEncoderH265 *const encoder = GST_VAAPI_ENCODER_H265 (base_encoder);
   GstVideoInfo *const vip = GST_VAAPI_ENCODER_VIDEO_INFO (encoder);
   const guint DEFAULT_SURFACES_COUNT = 3;
 
-  /* Fixme: Using only a rough approximation for bitstream headers..
-   * Fixme: Not taken into account: ScalingList, RefPicListModification, PredWeightTable */
+  /* FIXME: Using only a rough approximation for bitstream headers.
+   * Not taken into account: ScalingList, RefPicListModification,
+   * PredWeightTable */
   /* Maximum sizes for common headers (in bits) */
   enum
   {
@@ -2350,8 +2433,8 @@ set_context_info (GstVaapiEncoder * base_encoder)
   if (!ensure_hw_profile (encoder))
     return GST_VAAPI_ENCODER_STATUS_ERROR_UNSUPPORTED_PROFILE;
 
-  base_encoder->num_ref_frames =
-      ((encoder->num_bframes ? 2 : 1) + DEFAULT_SURFACES_COUNT);
+  base_encoder->num_ref_frames = (encoder->num_ref_frames
+      + (encoder->num_bframes > 0 ? 1 : 0) + DEFAULT_SURFACES_COUNT);
 
   /* Only YUV 4:2:0 formats are supported for now. */
   base_encoder->codedbuf_size += GST_ROUND_UP_32 (vip->width) *
@@ -2363,8 +2446,7 @@ set_context_info (GstVaapiEncoder * base_encoder)
 static GstVaapiEncoderStatus
 gst_vaapi_encoder_h265_reconfigure (GstVaapiEncoder * base_encoder)
 {
-  GstVaapiEncoderH265 *const encoder =
-      GST_VAAPI_ENCODER_H265_CAST (base_encoder);
+  GstVaapiEncoderH265 *const encoder = GST_VAAPI_ENCODER_H265 (base_encoder);
   GstVaapiEncoderStatus status;
   guint luma_width, luma_height;
 
@@ -2402,14 +2484,14 @@ gst_vaapi_encoder_h265_reconfigure (GstVaapiEncoder * base_encoder)
     return status;
 
   reset_properties (encoder);
+  ensure_control_rate_params (encoder);
   return set_context_info (base_encoder);
 }
 
 static gboolean
 gst_vaapi_encoder_h265_init (GstVaapiEncoder * base_encoder)
 {
-  GstVaapiEncoderH265 *const encoder =
-      GST_VAAPI_ENCODER_H265_CAST (base_encoder);
+  GstVaapiEncoderH265 *const encoder = GST_VAAPI_ENCODER_H265 (base_encoder);
   GstVaapiH265ReorderPool *reorder_pool;
   GstVaapiH265RefPool *ref_pool;
 
@@ -2437,8 +2519,7 @@ static void
 gst_vaapi_encoder_h265_finalize (GstVaapiEncoder * base_encoder)
 {
   /*free private buffers */
-  GstVaapiEncoderH265 *const encoder =
-      GST_VAAPI_ENCODER_H265_CAST (base_encoder);
+  GstVaapiEncoderH265 *const encoder = GST_VAAPI_ENCODER_H265 (base_encoder);
   GstVaapiEncPicture *pic;
   GstVaapiEncoderH265Ref *ref;
   GstVaapiH265RefPool *ref_pool;
@@ -2470,8 +2551,7 @@ static GstVaapiEncoderStatus
 gst_vaapi_encoder_h265_set_property (GstVaapiEncoder * base_encoder,
     gint prop_id, const GValue * value)
 {
-  GstVaapiEncoderH265 *const encoder =
-      GST_VAAPI_ENCODER_H265_CAST (base_encoder);
+  GstVaapiEncoderH265 *const encoder = GST_VAAPI_ENCODER_H265 (base_encoder);
 
   switch (prop_id) {
     case GST_VAAPI_ENCODER_H265_PROP_MAX_BFRAMES:
@@ -2483,12 +2563,24 @@ gst_vaapi_encoder_h265_set_property (GstVaapiEncoder * base_encoder,
     case GST_VAAPI_ENCODER_H265_PROP_MIN_QP:
       encoder->min_qp = g_value_get_uint (value);
       break;
+    case GST_VAAPI_ENCODER_H265_PROP_QP_IP:
+      encoder->qp_ip = g_value_get_int (value);
+      break;
+    case GST_VAAPI_ENCODER_H265_PROP_QP_IB:
+      encoder->qp_ib = g_value_get_int (value);
+      break;
     case GST_VAAPI_ENCODER_H265_PROP_NUM_SLICES:
       encoder->num_slices = g_value_get_uint (value);
       break;
     case GST_VAAPI_ENCODER_H265_PROP_CPB_LENGTH:
       encoder->cpb_length = g_value_get_uint (value);
       break;
+    case GST_VAAPI_ENCODER_H265_PROP_NUM_REF_FRAMES:
+      encoder->num_ref_frames = g_value_get_uint (value);
+      break;
+    case GST_VAAPI_ENCODER_H265_PROP_MBBRC:
+      encoder->mbbrc = g_value_get_enum (value);
+      break;
     default:
       return GST_VAAPI_ENCODER_STATUS_ERROR_INVALID_PARAMETER;
   }
@@ -2555,6 +2647,18 @@ gst_vaapi_encoder_h265_get_default_properties (void)
           "Max B-Frames", "Number of B-frames between I and P", 0, 10, 0,
           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
+  /**
+   * GstVaapiEncoderH265:refs:
+   *
+   * The number of reference frames.
+   * If B frame is encoded, it will add 1 reference frame more.
+   */
+  GST_VAAPI_ENCODER_PROPERTIES_APPEND (props,
+      GST_VAAPI_ENCODER_H265_PROP_NUM_REF_FRAMES,
+      g_param_spec_uint ("refs",
+          "Number of Reference Frames", "Number of reference frames", 1, 3, 1,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
   /**
    * GstVaapiEncoderH265:init-qp:
    *
@@ -2577,7 +2681,33 @@ gst_vaapi_encoder_h265_get_default_properties (void)
           "Minimum QP", "Minimum quantizer value", 1, 51, 1,
           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
-  /* Fixme: there seems to be issues with multi-slice encoding */
+  /**
+   * GstVaapiEncoderH265:qp-ip:
+   *
+   * The difference of QP between I and P Frame.
+   * This is available only on CQP mode.
+   */
+  GST_VAAPI_ENCODER_PROPERTIES_APPEND (props,
+      GST_VAAPI_ENCODER_H265_PROP_QP_IP,
+      g_param_spec_int ("qp-ip",
+          "Difference of QP between I and P frame",
+          "Difference of QP between I and P frame (available only on CQP)",
+          -51, 51, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
+  /**
+   * GstVaapiEncoderH265:qp-ib:
+   *
+   * The difference of QP between I and B Frame.
+   * This is available only on CQP mode.
+   */
+  GST_VAAPI_ENCODER_PROPERTIES_APPEND (props,
+      GST_VAAPI_ENCODER_H265_PROP_QP_IB,
+      g_param_spec_int ("qp-ib",
+          "Difference of QP between I and B frame",
+          "Difference of QP between I and B frame (available only on CQP)",
+          -51, 51, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
+  /* FIXME: there seems to be issues with multi-slice encoding */
   /**
    * GstVaapiEncoderH265:num-slices:
    *
@@ -2602,6 +2732,20 @@ gst_vaapi_encoder_h265_get_default_properties (void)
           1, 10000, DEFAULT_CPB_LENGTH,
           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
+  /**
+   * GstVaapiEncoderH265:mbbrc:
+   *
+   * Macroblock level bitrate control.
+   * This is not compatible with Constant QP rate control.
+   */
+  GST_VAAPI_ENCODER_PROPERTIES_APPEND (props,
+      GST_VAAPI_ENCODER_H265_PROP_MBBRC,
+      g_param_spec_enum ("mbbrc",
+          "Macroblock level Bitrate Control",
+          "Macroblock level Bitrate Control",
+          GST_VAAPI_TYPE_ENCODER_MBBRC, GST_VAAPI_ENCODER_MBBRC_AUTO,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
   return props;
 }
 
