@@ -26,6 +26,7 @@
 #include <gst/vaapi/gstvaapidisplay.h>
 
 #include "gstvaapidecode.h"
+#include "gstvaapidecode_props.h"
 #include "gstvaapipluginutil.h"
 #include "gstvaapivideobuffer.h"
 #if (USE_GLX || USE_EGL)
@@ -86,7 +87,8 @@ static const char gst_vaapidecode_src_caps_str[] =
 #if (USE_GLX || USE_EGL)
     GST_VAAPI_MAKE_GLTEXUPLOAD_CAPS ";"
 #endif
-    GST_VIDEO_CAPS_MAKE("{ NV12, I420, YV12, P010_10LE }");
+    GST_VIDEO_CAPS_MAKE("{ NV12, I420, YV12, P010_10LE }") ";"
+    GST_VAAPI_MAKE_DMABUF_CAPS;
 
 static GstStaticPadTemplate gst_vaapidecode_src_factory =
     GST_STATIC_PAD_TEMPLATE(
@@ -103,31 +105,34 @@ struct _GstVaapiDecoderMap
   guint rank;
   const gchar *name;
   const gchar *caps_str;
+
+  void (*install_properties) (GObjectClass * klass);
 };
 
 static const GstVaapiDecoderMap vaapi_decode_map[] = {
 #if USE_JPEG_DECODER
-  {GST_VAAPI_CODEC_JPEG, GST_RANK_MARGINAL, "jpeg", "image/jpeg"},
+  {GST_VAAPI_CODEC_JPEG, GST_RANK_MARGINAL, "jpeg", "image/jpeg", NULL},
 #endif
   {GST_VAAPI_CODEC_MPEG2, GST_RANK_PRIMARY, "mpeg2",
-      "video/mpeg, mpegversion=2, systemstream=(boolean)false"},
+      "video/mpeg, mpegversion=2, systemstream=(boolean)false", NULL},
   {GST_VAAPI_CODEC_MPEG4, GST_RANK_PRIMARY, "mpeg4",
-      "video/mpeg, mpegversion=4"},
-  {GST_VAAPI_CODEC_H263, GST_RANK_PRIMARY, "h263", "video/x-h263"},
-  {GST_VAAPI_CODEC_H264, GST_RANK_PRIMARY, "h264", "video/x-h264"},
+      "video/mpeg, mpegversion=4", NULL},
+  {GST_VAAPI_CODEC_H263, GST_RANK_PRIMARY, "h263", "video/x-h263", NULL},
+  {GST_VAAPI_CODEC_H264, GST_RANK_PRIMARY, "h264", "video/x-h264",
+      gst_vaapi_decode_h264_install_properties},
   {GST_VAAPI_CODEC_VC1, GST_RANK_PRIMARY, "vc1",
-      "video/x-wmv, wmvversion=3, format={WMV3,WVC1}"},
+      "video/x-wmv, wmvversion=3, format={WMV3,WVC1}", NULL},
 #if USE_VP8_DECODER
-  {GST_VAAPI_CODEC_VP8, GST_RANK_PRIMARY, "vp8", "video/x-vp8"},
+  {GST_VAAPI_CODEC_VP8, GST_RANK_PRIMARY, "vp8", "video/x-vp8", NULL},
 #endif
 #if USE_VP9_DECODER
-  {GST_VAAPI_CODEC_VP9, GST_RANK_PRIMARY, "vp9", "video/x-vp9"},
+  {GST_VAAPI_CODEC_VP9, GST_RANK_PRIMARY, "vp9", "video/x-vp9", NULL},
 #endif
 #if USE_H265_DECODER
-  {GST_VAAPI_CODEC_H265, GST_RANK_PRIMARY, "h265", "video/x-h265"},
+  {GST_VAAPI_CODEC_H265, GST_RANK_PRIMARY, "h265", "video/x-h265", NULL},
 #endif
   {0 /* the rest */ , GST_RANK_PRIMARY + 1, NULL,
-      gst_vaapidecode_sink_caps_str},
+      gst_vaapidecode_sink_caps_str, NULL},
 };
 
 static GstElementClass *parent_class = NULL;
@@ -230,6 +235,8 @@ gst_vaapidecode_ensure_allowed_srcpad_caps (GstVaapiDecode * decode)
         gst_caps_from_string (GST_VAAPI_MAKE_GLTEXUPLOAD_CAPS));
   }
 #endif
+  out_caps = gst_caps_make_writable (out_caps);
+  gst_caps_append (out_caps, gst_caps_from_string (GST_VAAPI_MAKE_DMABUF_CAPS));
 
   raw_caps = gst_vaapi_plugin_base_get_allowed_raw_caps
       (GST_VAAPI_PLUGIN_BASE (decode));
@@ -327,6 +334,7 @@ gst_vaapidecode_update_src_caps (GstVaapiDecode * decode)
 
   switch (feature) {
     case GST_VAAPI_CAPS_FEATURE_GL_TEXTURE_UPLOAD_META:
+    case GST_VAAPI_CAPS_FEATURE_DMABUF:
     case GST_VAAPI_CAPS_FEATURE_VAAPI_SURFACE:{
       GstStructure *structure = gst_caps_get_structure (state->caps, 0);
 
@@ -844,6 +852,8 @@ gst_vaapidecode_create (GstVaapiDecode * decode, GstCaps * caps)
 
       /* Set the stream buffer alignment for better optimizations */
       if (decode->decoder && caps) {
+        GstVaapiDecodeH264Private *priv =
+            gst_vaapi_decode_h264_get_instance_private (decode);
         GstStructure *const structure = gst_caps_get_structure (caps, 0);
         const gchar *str = NULL;
 
@@ -858,6 +868,13 @@ gst_vaapidecode_create (GstVaapiDecode * decode, GstCaps * caps)
           gst_vaapi_decoder_h264_set_alignment (GST_VAAPI_DECODER_H264
               (decode->decoder), alignment);
         }
+
+        if (priv) {
+          gst_vaapi_decoder_h264_set_low_latency (GST_VAAPI_DECODER_H264
+              (decode->decoder), priv->is_low_latency);
+          gst_vaapi_decoder_h264_set_base_only (GST_VAAPI_DECODER_H264
+              (decode->decoder), priv->base_only);
+        }
       }
       break;
 #if USE_H265_DECODER
@@ -1139,12 +1156,40 @@ gst_vaapidecode_parse (GstVideoDecoder * vdec,
   return ret;
 }
 
+static gboolean
+is_mvc_profile (GstVaapiProfile profile)
+{
+  return profile == GST_VAAPI_PROFILE_H264_MULTIVIEW_HIGH
+      || profile == GST_VAAPI_PROFILE_H264_STEREO_HIGH;
+}
+
+static gboolean
+is_svc_profile (GstVaapiProfile profile)
+{
+  return profile == GST_VAAPI_PROFILE_H264_SCALABLE_BASELINE
+      || profile == GST_VAAPI_PROFILE_H264_SCALABLE_HIGH;
+}
+
+
+static GstCaps *
+add_h264_profile_in_caps (GstCaps * caps, const gchar * profile_name)
+{
+  GstCaps *caps_new =
+      gst_caps_new_simple ("video/x-h264", "profile", G_TYPE_STRING,
+      profile_name, NULL);
+  return gst_caps_merge (caps_new, caps);
+}
+
 static gboolean
 gst_vaapidecode_ensure_allowed_sinkpad_caps (GstVaapiDecode * decode)
 {
   GstCaps *caps, *allowed_sinkpad_caps;
   GArray *profiles;
   guint i;
+  gboolean base_only = FALSE;
+  gboolean have_high = FALSE;
+  gboolean have_mvc = FALSE;
+  gboolean have_svc = FALSE;
 
   profiles =
       gst_vaapi_display_get_decode_profiles (GST_VAAPI_PLUGIN_BASE_DISPLAY
@@ -1156,6 +1201,10 @@ gst_vaapidecode_ensure_allowed_sinkpad_caps (GstVaapiDecode * decode)
   if (!allowed_sinkpad_caps)
     goto error_no_memory;
 
+  if (g_object_class_find_property (G_OBJECT_GET_CLASS (decode), "base-only")) {
+    g_object_get (decode, "base-only", &base_only, NULL);
+  }
+
   for (i = 0; i < profiles->len; i++) {
     const GstVaapiProfile profile =
         g_array_index (profiles, GstVaapiProfile, i);
@@ -1178,6 +1227,38 @@ gst_vaapidecode_ensure_allowed_sinkpad_caps (GstVaapiDecode * decode)
           profile_name, NULL);
 
     allowed_sinkpad_caps = gst_caps_merge (allowed_sinkpad_caps, caps);
+    have_mvc |= is_mvc_profile (profile);
+    have_svc |= is_svc_profile (profile);
+    have_high |= profile == GST_VAAPI_PROFILE_H264_HIGH;
+  }
+
+  if (base_only && (!have_mvc || !have_svc) && have_high) {
+    if (!have_mvc) {
+      GST_DEBUG ("base_only: force adding MVC profiles in caps");
+
+      allowed_sinkpad_caps =
+          add_h264_profile_in_caps (allowed_sinkpad_caps, "multiview-high");
+      allowed_sinkpad_caps =
+          add_h264_profile_in_caps (allowed_sinkpad_caps, "stereo-high");
+    }
+
+    if (!have_svc) {
+      GST_DEBUG ("base_only: force adding SVC profiles in caps");
+
+      allowed_sinkpad_caps =
+          add_h264_profile_in_caps (allowed_sinkpad_caps,
+          "scalable-constrained-baseline");
+      allowed_sinkpad_caps =
+          add_h264_profile_in_caps (allowed_sinkpad_caps, "scalable-baseline");
+      allowed_sinkpad_caps =
+          add_h264_profile_in_caps (allowed_sinkpad_caps,
+          "scalable-high-intra");
+      allowed_sinkpad_caps =
+          add_h264_profile_in_caps (allowed_sinkpad_caps,
+          "scalable-constrained-high");
+      allowed_sinkpad_caps =
+          add_h264_profile_in_caps (allowed_sinkpad_caps, "scalable-high");
+    }
   }
   decode->allowed_sinkpad_caps = gst_caps_simplify (allowed_sinkpad_caps);
 
@@ -1364,6 +1445,9 @@ gst_vaapidecode_class_init (GstVaapiDecodeClass * klass)
   g_free (longname);
   g_free (description);
 
+  if (map->install_properties)
+    map->install_properties (object_class);
+
   /* sink pad */
   caps = gst_caps_from_string (map->caps_str);
   pad_template = gst_pad_template_new ("sink", GST_PAD_SINK, GST_PAD_ALWAYS,
