@@ -175,11 +175,18 @@ gst_vaapiencode_default_set_property (GstVaapiEncode * encode, guint prop_id,
     const GValue * value)
 {
   PropValue *const prop_value = prop_value_lookup (encode, prop_id);
+  GstVaapiEncoder *encoder = encode->encoder;
 
   if (prop_value) {
     g_value_copy (value, &prop_value->value);
+
+    if (encoder)
+      return (gst_vaapi_encoder_set_property (encoder, prop_id,
+              value) == GST_VAAPI_ENCODER_STATUS_SUCCESS);
+
     return TRUE;
   }
+
   return FALSE;
 }
 
@@ -271,6 +278,9 @@ gst_vaapiencode_push_frame (GstVaapiEncode * encode, gint64 timeout)
   GstVaapiEncoderStatus status;
   GstBuffer *out_buffer;
   GstFlowReturn ret;
+#if USE_H264_FEI_ENCODER
+  GstVaapiFeiVideoMeta *feimeta = NULL;
+#endif
 
   status = gst_vaapi_encoder_get_buffer_with_timeout (encode->encoder,
       &codedbuf_proxy, timeout);
@@ -295,6 +305,15 @@ gst_vaapiencode_push_frame (GstVaapiEncode * encode, gint64 timeout)
   out_buffer = NULL;
   ret = klass->alloc_buffer (encode,
       GST_VAAPI_CODED_BUFFER_PROXY_BUFFER (codedbuf_proxy), &out_buffer);
+
+#if USE_H264_FEI_ENCODER
+  if (klass->save_stats_to_meta) {
+    feimeta = klass->save_stats_to_meta (encode, codedbuf_proxy);
+    if (feimeta != NULL)
+      gst_buffer_set_vaapi_fei_video_meta (out_buffer, feimeta);
+  }
+#endif
+
   gst_vaapi_coded_buffer_proxy_replace (&codedbuf_proxy, NULL);
   if (ret != GST_FLOW_OK)
     goto error_allocate_buffer;
@@ -636,6 +655,10 @@ gst_vaapiencode_handle_frame (GstVideoEncoder * venc,
   GstVaapiSurfaceProxy *proxy;
   GstFlowReturn ret;
   GstBuffer *buf;
+#if USE_H264_FEI_ENCODER
+  GstVaapiFeiVideoMeta *feimeta = NULL;
+  GstVaapiEncodeClass *const klass = GST_VAAPIENCODE_GET_CLASS (venc);
+#endif
 
   buf = NULL;
   ret = gst_vaapi_plugin_base_get_input_buffer (GST_VAAPI_PLUGIN_BASE (encode),
@@ -654,6 +677,12 @@ gst_vaapiencode_handle_frame (GstVideoEncoder * venc,
   if (!proxy)
     goto error_buffer_no_surface_proxy;
 
+#if USE_H264_FEI_ENCODER
+  feimeta = gst_buffer_get_vaapi_fei_video_meta (buf);
+  if (feimeta && klass->load_control_data)
+    klass->load_control_data (encode, feimeta, proxy);
+#endif
+
   gst_video_codec_frame_set_user_data (frame,
       gst_vaapi_surface_proxy_ref (proxy),
       (GDestroyNotify) gst_vaapi_surface_proxy_unref);
@@ -758,6 +787,45 @@ gst_vaapiencode_sink_event (GstVideoEncoder * venc, GstEvent * event)
   GstPad *const srcpad = GST_VAAPI_PLUGIN_BASE_SRC_PAD (encode);
   gboolean ret;
 
+  switch (GST_EVENT_TYPE (event)) {
+    case GST_EVENT_CUSTOM_DOWNSTREAM:{
+      const GstStructure *s = gst_event_get_structure (event);
+      if (gst_structure_has_name (s, "GstVaapiEncoderRegionOfInterest")) {
+        GstVaapiROI roi;
+
+        if (!encode->encoder)
+          return TRUE;
+
+        if (!gst_structure_get_uint (s, "roi-x", &roi.rect.x) ||
+            !gst_structure_get_uint (s, "roi-y", &roi.rect.y) ||
+            !gst_structure_get_uint (s, "roi-width", &roi.rect.width) ||
+            !gst_structure_get_uint (s, "roi-height", &roi.rect.height) ||
+            !gst_structure_get_int (s, "roi-value", &roi.roi_value)) {
+          return TRUE;
+        }
+
+        if (roi.roi_value == 0) {
+          ret = gst_vaapi_encoder_del_roi (encode->encoder, &roi);
+          if (ret) {
+            GST_INFO_OBJECT (venc, "ROI: region with %d/%d/%d/%d is removed",
+                roi.rect.x, roi.rect.y, roi.rect.width, roi.rect.height);
+          }
+        } else {
+          ret = gst_vaapi_encoder_add_roi (encode->encoder, &roi);
+          if (ret) {
+            GST_INFO_OBJECT (venc, "ROI: region with %d/%d/%d/%d is added",
+                roi.rect.x, roi.rect.y, roi.rect.width, roi.rect.height);
+          }
+        }
+        gst_event_unref (event);
+        return ret;
+      }
+      break;
+    }
+    default:
+      break;
+  }
+
   ret = GST_VIDEO_ENCODER_CLASS (gst_vaapiencode_parent_class)->sink_event
       (venc, event);
   if (!ret)
