@@ -59,6 +59,21 @@ plugin_set_display (GstVaapiPluginBase * plugin, GstVaapiDisplay * display)
   gst_vaapi_display_unref (display);
 }
 
+#if USE_GST_GL_HELPERS
+static void
+plugin_set_gst_gl (GstVaapiPluginBase * plugin, GstGLDisplay * gl_display,
+    GstGLContext * gl_context, GstGLContext * gl_other_context)
+{
+  gst_object_replace (&plugin->gl_display, NULL);
+  plugin->gl_display = (GstObject *) gl_display;
+
+  gst_object_replace (&plugin->gl_context, (GstObject *) gl_context);
+
+  gst_object_replace (&plugin->gl_other_context, NULL);
+  plugin->gl_other_context = (GstObject *) gl_other_context;
+}
+#endif
+
 /**
  * gst_vaapi_plugin_base_set_context:
  * @plugin: a #GstVaapiPluginBase instance
@@ -76,13 +91,21 @@ gst_vaapi_plugin_base_set_context (GstVaapiPluginBase * plugin,
 {
   GstVaapiDisplay *display = NULL;
 
-  if (gst_vaapi_video_context_get_display (context, &display))
+  /* gst.vaapi.app.Display is only attended _if_ the element is
+   * vaapisink and it doesn't have a display set yet */
+  if (gst_vaapi_video_context_get_display (context,
+          GST_IS_VIDEO_SINK (plugin) && !plugin->display, &display)) {
     plugin_set_display (plugin, display);
-
+  }
 #if USE_GST_GL_HELPERS
-  gst_gl_handle_set_context (GST_ELEMENT_CAST (plugin), context,
-      (GstGLDisplay **) & plugin->gl_display,
-      (GstGLContext **) & plugin->gl_other_context);
+  {
+    GstGLDisplay *gl_display = NULL;
+    GstGLContext *gl_other_context = NULL;
+    GstElement *el = GST_ELEMENT_CAST (plugin);
+
+    if (gst_gl_handle_set_context (el, context, &gl_display, &gl_other_context))
+      plugin_set_gst_gl (plugin, gl_display, NULL, gl_other_context);
+  }
 #endif
 }
 
@@ -298,10 +321,8 @@ gst_vaapi_plugin_base_close (GstVaapiPluginBase * plugin)
 
   gst_caps_replace (&plugin->sinkpad_caps, NULL);
   gst_video_info_init (&plugin->sinkpad_info);
-  if (plugin->sinkpad_buffer_pool) {
-    gst_object_unref (plugin->sinkpad_buffer_pool);
-    plugin->sinkpad_buffer_pool = NULL;
-  }
+
+  g_clear_object (&plugin->sinkpad_buffer_pool);
   g_clear_object (&plugin->srcpad_buffer_pool);
 
   g_clear_object (&plugin->sinkpad_allocator);
@@ -398,68 +419,6 @@ gst_vaapi_plugin_base_ensure_display (GstVaapiPluginBase * plugin)
   return TRUE;
 }
 
-/* Checks whether the supplied pad peer element supports DMABUF sharing */
-/* XXX: this is a workaround to the absence of any proposer way to
-   specify DMABUF memory capsfeatures or bufferpool option to downstream */
-static gboolean
-has_dmabuf_capable_peer (GstVaapiPluginBase * plugin, GstPad * pad)
-{
-  GstPad *other_pad = NULL;
-  GstElement *element = NULL;
-  gchar *element_name = NULL;
-  gboolean is_dmabuf_capable = FALSE;
-  gint v;
-
-  gst_object_ref (pad);
-
-  for (;;) {
-    other_pad = gst_pad_get_peer (pad);
-    gst_object_unref (pad);
-    if (!other_pad)
-      break;
-
-    element = gst_pad_get_parent_element (other_pad);
-    gst_object_unref (other_pad);
-    if (!element)
-      break;
-
-    if (GST_IS_PUSH_SRC (element)) {
-      element_name = gst_element_get_name (element);
-      if (!element_name)
-        break;
-
-      if ((sscanf (element_name, "v4l2src%d", &v) != 1)
-          && (sscanf (element_name, "camerasrc%d", &v) != 1))
-        break;
-
-      v = 0;
-      g_object_get (element, "io-mode", &v, NULL);
-      if (strncmp (element_name, "camerasrc", 9) == 0)
-        is_dmabuf_capable = v == 3;
-      else
-        is_dmabuf_capable = v == 5;     /* "dmabuf-import" enum value */
-      break;
-    } else if (GST_IS_BASE_TRANSFORM (element)) {
-      element_name = gst_element_get_name (element);
-      if (!element_name || sscanf (element_name, "capsfilter%d", &v) != 1)
-        break;
-
-      pad = gst_element_get_static_pad (element, "sink");
-      if (!pad)
-        break;
-    } else
-      break;
-
-    g_free (element_name);
-    element_name = NULL;
-    g_clear_object (&element);
-  }
-
-  g_free (element_name);
-  g_clear_object (&element);
-  return is_dmabuf_capable;
-}
-
 static gboolean
 gst_vaapi_buffer_pool_caps_is_equal (GstBufferPool * pool, GstCaps * newcaps)
 {
@@ -498,23 +457,15 @@ ensure_sinkpad_allocator (GstVaapiPluginBase * plugin, GstCaps * caps,
     guint * size)
 {
   GstVideoInfo vinfo;
+  const GstVideoInfo *image_info;
   GstVaapiImageUsageFlags usage_flag =
       GST_VAAPI_IMAGE_USAGE_FLAG_NATIVE_FORMATS;
 
   if (!gst_video_info_from_caps (&vinfo, caps))
     goto error_invalid_caps;
-  gst_video_info_force_nv12_if_encoded (&vinfo);
-  *size = GST_VIDEO_INFO_SIZE (&vinfo);
 
   if (!reset_allocator (plugin->sinkpad_allocator, &vinfo))
-    return TRUE;
-
-  if (has_dmabuf_capable_peer (plugin, plugin->sinkpad)) {
-    plugin->sinkpad_allocator =
-        gst_vaapi_dmabuf_allocator_new (plugin->display, &vinfo,
-        GST_VAAPI_SURFACE_ALLOC_FLAG_LINEAR_STORAGE, GST_PAD_SINK);
     goto bail;
-  }
 
   /* enable direct upload if upstream requests raw video */
   if (gst_caps_is_video_raw (caps)) {
@@ -527,6 +478,14 @@ ensure_sinkpad_allocator (GstVaapiPluginBase * plugin, GstCaps * caps,
 bail:
   if (!plugin->sinkpad_allocator)
     goto error_create_allocator;
+
+  image_info =
+      gst_allocator_get_vaapi_video_info (plugin->sinkpad_allocator, NULL);
+  g_assert (image_info);        /* allocator ought set its image info */
+
+  /* update the size with the one generated by the allocator */
+  *size = GST_VIDEO_INFO_SIZE (image_info);
+
   return TRUE;
 
   /* ERRORS */
@@ -562,14 +521,78 @@ get_dmabuf_surface_allocation_flags (void)
   return GST_VAAPI_SURFACE_ALLOC_FLAG_LINEAR_STORAGE;
 }
 
+static inline GstAllocator *
+create_dmabuf_srcpad_allocator (GstVaapiPluginBase * plugin,
+    GstVideoInfo * vinfo, gboolean check_for_map)
+{
+  GstAllocator *allocator;
+
+  if (!GST_IS_VIDEO_DECODER (plugin) && !GST_IS_BASE_TRANSFORM (plugin))
+    return NULL;
+
+  allocator = gst_vaapi_dmabuf_allocator_new (plugin->display, vinfo,
+      get_dmabuf_surface_allocation_flags (), GST_PAD_SRC);
+  if (!allocator || !check_for_map)
+    return allocator;
+
+  /* the dmabuf allocator *must* be capable to map a buffer with raw
+   * caps and the there's no evidence of downstream dmabuf
+   * importation */
+  if (!gst_vaapi_dmabuf_can_map (plugin->display, allocator)) {
+    GST_INFO_OBJECT (plugin, "dmabuf allocator generates unmappable buffers");
+    gst_object_replace ((GstObject **) & allocator, NULL);
+  }
+
+  return allocator;
+}
+
 static gboolean
 ensure_srcpad_allocator (GstVaapiPluginBase * plugin, GstVideoInfo * vinfo,
     GstCaps * caps)
 {
   gboolean different_caps;
-  GstVideoInfo vi;
-  GstVaapiImageUsageFlags usage_flag =
-      GST_VAAPI_IMAGE_USAGE_FLAG_NATIVE_FORMATS;
+  const GstVideoInfo *image_info;
+
+  if (!reset_allocator (plugin->srcpad_allocator, vinfo))
+    goto valid_allocator;
+
+  plugin->srcpad_allocator = NULL;
+  if (caps && gst_caps_is_video_raw (caps)) {
+    GstAllocator *allocator = create_dmabuf_srcpad_allocator (plugin, vinfo,
+        !plugin->srcpad_can_dmabuf);
+    plugin->srcpad_allocator = allocator;
+  } else if (caps && gst_vaapi_caps_feature_contains (caps,
+          GST_VAAPI_CAPS_FEATURE_DMABUF)) {
+    plugin->srcpad_allocator =
+        create_dmabuf_srcpad_allocator (plugin, vinfo, FALSE);
+    if (!plugin->srcpad_allocator)
+      goto error_create_allocator;
+  }
+
+  if (!plugin->srcpad_allocator) {
+    GstVaapiImageUsageFlags usage_flag =
+        GST_VAAPI_IMAGE_USAGE_FLAG_NATIVE_FORMATS;
+
+    if (plugin->enable_direct_rendering) {
+      usage_flag = GST_VAAPI_IMAGE_USAGE_FLAG_DIRECT_RENDER;
+      GST_INFO_OBJECT (plugin, "enabling direct rendering in source allocator");
+    }
+
+    plugin->srcpad_allocator =
+        gst_vaapi_video_allocator_new (plugin->display, vinfo, 0, usage_flag);
+  }
+
+  if (!plugin->srcpad_allocator)
+    goto error_create_allocator;
+
+valid_allocator:
+  image_info =
+      gst_allocator_get_vaapi_video_info (plugin->srcpad_allocator, NULL);
+  g_assert (image_info);        /* both allocators ought set its image
+                                 * info */
+
+  /* update the size with the one generated by the allocator */
+  GST_VIDEO_INFO_SIZE (vinfo) = GST_VIDEO_INFO_SIZE (image_info);
 
   /* the received caps are the "allocation caps" which may be
    * different from the "negotiation caps". In this case, we should
@@ -579,52 +602,20 @@ ensure_srcpad_allocator (GstVaapiPluginBase * plugin, GstVideoInfo * vinfo,
       !gst_caps_is_strictly_equal (plugin->srcpad_caps, caps);
 
   if (different_caps) {
-    vi = plugin->srcpad_info;
-  } else {
-    vi = *vinfo;
-  }
+    guint i;
+    GstVideoInfo vi = plugin->srcpad_info;
 
-  if (!reset_allocator (plugin->srcpad_allocator, &vi))
-    return TRUE;
-
-  plugin->srcpad_allocator = NULL;
-  if (caps && gst_caps_is_video_raw (caps)) {
-    if (plugin->srcpad_can_dmabuf) {
-      if (GST_IS_VIDEO_DECODER (plugin) || GST_IS_BASE_TRANSFORM (plugin)) {
-        plugin->srcpad_allocator =
-            gst_vaapi_dmabuf_allocator_new (plugin->display, vinfo,
-            get_dmabuf_surface_allocation_flags (), GST_PAD_SRC);
-      }
-    } else if (plugin->enable_direct_rendering) {
-      usage_flag = GST_VAAPI_IMAGE_USAGE_FLAG_DIRECT_RENDER;
-      GST_INFO_OBJECT (plugin, "enabling direct rendering in source allocator");
-    }
-  }
-
-  if (!plugin->srcpad_allocator) {
-    plugin->srcpad_allocator =
-        gst_vaapi_video_allocator_new (plugin->display, vinfo, 0, usage_flag);
-  }
-
-  if (!plugin->srcpad_allocator)
-    goto error_create_allocator;
-
-  if (different_caps) {
-    guint i, flags = 0;
-    const GstVideoInfo *alloc_vi =
-        gst_allocator_get_vaapi_video_info (plugin->srcpad_allocator, &flags);
-    /* update the planes and the size with the allocator image info,
-     * but not the resolution */
-    if (alloc_vi) {
-      for (i = 0; i < GST_VIDEO_INFO_N_PLANES (alloc_vi); i++) {
-        GST_VIDEO_INFO_PLANE_OFFSET (&vi, i) =
-            GST_VIDEO_INFO_PLANE_OFFSET (alloc_vi, i);
-        GST_VIDEO_INFO_PLANE_STRIDE (&vi, i) =
-            GST_VIDEO_INFO_PLANE_STRIDE (alloc_vi, i);
-      }
-      GST_VIDEO_INFO_SIZE (&vi) = GST_VIDEO_INFO_SIZE (alloc_vi);
-      gst_allocator_set_vaapi_video_info (plugin->srcpad_allocator, &vi, flags);
+    /* update the planes and the size with the allocator image/surface
+     * info, but not the resolution */
+    for (i = 0; i < GST_VIDEO_INFO_N_PLANES (image_info); i++) {
+      GST_VIDEO_INFO_PLANE_OFFSET (&vi, i) =
+          GST_VIDEO_INFO_PLANE_OFFSET (image_info, i);
+      GST_VIDEO_INFO_PLANE_STRIDE (&vi, i) =
+          GST_VIDEO_INFO_PLANE_STRIDE (image_info, i);
     }
+    GST_VIDEO_INFO_SIZE (&vi) = GST_VIDEO_INFO_SIZE (image_info);
+    gst_allocator_set_vaapi_negotiated_video_info (plugin->srcpad_allocator,
+        &vi);
   }
 
   return TRUE;
@@ -813,20 +804,46 @@ gst_vaapi_plugin_base_propose_allocation (GstVaapiPluginBase * plugin,
     GstQuery * query)
 {
   GstCaps *caps = NULL;
+  GstBufferPool *pool = NULL;
   gboolean need_pool;
+  guint size = 0, n_allocators;
 
   gst_query_parse_allocation (query, &caps, &need_pool);
   if (!caps)
     goto error_no_caps;
 
+  if (!ensure_sinkpad_allocator (plugin, caps, &size))
+    return FALSE;
+
   if (need_pool) {
-    if (!ensure_sinkpad_buffer_pool (plugin, caps))
+    pool = gst_vaapi_plugin_base_create_pool (plugin, caps, size,
+        BUFFER_POOL_SINK_MIN_BUFFERS, 0,
+        GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_VIDEO_META,
+        plugin->sinkpad_allocator);
+    if (!pool)
       return FALSE;
-    gst_query_add_allocation_pool (query, plugin->sinkpad_buffer_pool,
-        plugin->sinkpad_buffer_size, BUFFER_POOL_SINK_MIN_BUFFERS, 0);
-    gst_query_add_allocation_param (query, plugin->sinkpad_allocator, NULL);
   }
 
+  /* Set sinkpad allocator as the last allocation param.
+   *
+   * If there's none, set system's allocator first and VAAPI allocator
+   * second
+   */
+  n_allocators = gst_query_get_n_allocation_params (query);
+  if (n_allocators == 0) {
+    GstAllocator *allocator;
+
+    allocator = gst_allocator_find (GST_ALLOCATOR_SYSMEM);
+    gst_query_add_allocation_param (query, allocator, NULL);
+    gst_object_unref (allocator);
+  }
+  gst_query_add_allocation_param (query, plugin->sinkpad_allocator, NULL);
+
+  gst_query_add_allocation_pool (query, pool, size,
+      BUFFER_POOL_SINK_MIN_BUFFERS, 0);
+  if (pool)
+    gst_object_unref (pool);
+
   gst_query_add_allocation_meta (query, GST_VAAPI_VIDEO_META_API_TYPE, NULL);
   gst_query_add_allocation_meta (query, GST_VIDEO_META_API_TYPE, NULL);
   return TRUE;
@@ -857,8 +874,9 @@ gst_vaapi_plugin_base_decide_allocation (GstVaapiPluginBase * plugin,
   GstCaps *caps = NULL;
   GstBufferPool *pool;
   GstVideoInfo vi;
-  guint size, min, max, pool_options;
-  gboolean update_pool = FALSE, update_allocator = FALSE;
+  guint i, size, min, max, pool_options, num_allocators;
+  gint index_allocator;
+  gboolean update_pool = FALSE;
 #if (USE_GLX || USE_EGL)
   guint idx;
 #endif
@@ -889,6 +907,7 @@ gst_vaapi_plugin_base_decide_allocation (GstVaapiPluginBase * plugin,
       if (gst_structure_get (params, "gst.gl.GstGLContext", GST_TYPE_GL_CONTEXT,
               &gl_context, NULL) && gl_context) {
         gst_vaapi_plugin_base_set_gl_context (plugin, gl_context);
+        gst_vaapi_plugin_base_set_srcpad_can_dmabuf (plugin, gl_context);
         gst_object_unref (gl_context);
       }
     }
@@ -906,8 +925,25 @@ gst_vaapi_plugin_base_decide_allocation (GstVaapiPluginBase * plugin,
     goto error_invalid_caps;
   gst_video_info_force_nv12_if_encoded (&vi);
 
-  if (gst_query_get_n_allocation_params (query) > 0)
-    update_allocator = TRUE;
+  index_allocator = -1;
+  num_allocators = gst_query_get_n_allocation_params (query);
+  for (i = 0; i < num_allocators; i++) {
+    GstAllocator *allocator = NULL;
+
+    gst_query_parse_nth_allocation_param (query, i, &allocator, NULL);
+    if (!allocator)
+      continue;
+    if (g_strcmp0 (allocator->mem_type, GST_VAAPI_VIDEO_MEMORY_NAME) == 0) {
+      GST_DEBUG_OBJECT (plugin, "found vaapi allocator in query %"
+          GST_PTR_FORMAT, allocator);
+      index_allocator = i;
+      if (plugin->srcpad_allocator)
+        gst_object_unref (plugin->srcpad_allocator);
+      plugin->srcpad_allocator = allocator;
+      break;
+    }
+    gst_object_unref (allocator);
+  }
 
   if (gst_query_get_n_allocation_pools (query) > 0) {
     gst_query_parse_nth_allocation_pool (query, 0, &pool, &size, &min, &max);
@@ -937,6 +973,8 @@ gst_vaapi_plugin_base_decide_allocation (GstVaapiPluginBase * plugin,
   if (!pool) {
     if (!ensure_srcpad_allocator (plugin, &vi, caps))
       goto error;
+    size = GST_VIDEO_INFO_SIZE (&vi);   /* size might be updated by
+                                         * allocator */
     pool = gst_vaapi_plugin_base_create_pool (plugin, caps, size, min, max,
         pool_options, plugin->srcpad_allocator);
     if (!pool)
@@ -948,11 +986,17 @@ gst_vaapi_plugin_base_decide_allocation (GstVaapiPluginBase * plugin,
   else
     gst_query_add_allocation_pool (query, pool, size, min, max);
 
-  if (update_allocator)
-    gst_query_set_nth_allocation_param (query, 0, plugin->srcpad_allocator,
-        NULL);
-  else
-    gst_query_add_allocation_param (query, plugin->srcpad_allocator, NULL);
+  /* allocator might be updated by ensure_srcpad_allocator() */
+  if (plugin->srcpad_allocator) {
+    if (index_allocator > 0) {
+      gst_query_set_nth_allocation_param (query, index_allocator,
+          plugin->srcpad_allocator, NULL);
+    } else {
+      GST_DEBUG_OBJECT (plugin, "adding allocator in query %" GST_PTR_FORMAT,
+          plugin->srcpad_allocator);
+      gst_query_add_allocation_param (query, plugin->srcpad_allocator, NULL);
+    }
+  }
 
   g_clear_object (&plugin->srcpad_buffer_pool);
   plugin->srcpad_buffer_pool = pool;
@@ -1138,12 +1182,6 @@ gst_vaapi_plugin_base_set_gl_context (GstVaapiPluginBase * plugin,
       break;
 #endif
     case GST_GL_PLATFORM_EGL:
-#if VA_CHECK_VERSION (0,36,0) && USE_GST_GL_HELPERS
-      plugin->srcpad_can_dmabuf =
-          (!(gst_gl_context_get_gl_api (gl_context) & GST_GL_API_GLES1)
-          && gst_gl_context_check_feature (gl_context,
-              "EGL_EXT_image_dma_buf_import"));
-#endif
 #if USE_EGL
       display_type = GST_VAAPI_DISPLAY_TYPE_EGL;
       break;
@@ -1171,18 +1209,14 @@ GstObject *
 gst_vaapi_plugin_base_create_gl_context (GstVaapiPluginBase * plugin)
 {
 #if USE_GST_GL_HELPERS
-  GstGLContext *gl_other_context, *gl_context = NULL;
-  GstGLDisplay *gl_display;
+  GstGLContext *gl_other_context = NULL, *gl_context = NULL;
+  GstGLDisplay *gl_display = NULL;
 
-  if (!gst_gl_ensure_element_data (plugin,
-          (GstGLDisplay **) & plugin->gl_display,
-          (GstGLContext **) & plugin->gl_other_context))
+  if (!gst_gl_ensure_element_data (plugin, &gl_display, &gl_other_context))
     goto no_valid_gl_display;
 
-  gl_display = (GstGLDisplay *) plugin->gl_display;
   if (gst_gl_display_get_handle_type (gl_display) == GST_GL_DISPLAY_TYPE_ANY)
     goto no_valid_gl_display;
-  gl_other_context = (GstGLContext *) plugin->gl_other_context;
 
   GST_INFO_OBJECT (plugin, "creating a new GstGL context");
 
@@ -1199,13 +1233,15 @@ gst_vaapi_plugin_base_create_gl_context (GstVaapiPluginBase * plugin)
   } while (!gst_gl_display_add_context (gl_display, gl_context));
   GST_OBJECT_UNLOCK (gl_display);
 
+  plugin_set_gst_gl (plugin, gl_display, gl_context, gl_other_context);
+
   return GST_OBJECT_CAST (gl_context);
 
   /* ERRORS */
 no_valid_gl_display:
   {
-    gst_object_replace (&plugin->gl_display, NULL);
-    gst_object_replace (&plugin->gl_other_context, NULL);
+    gst_object_replace ((GstObject **) & gl_display, NULL);
+    gst_object_replace ((GstObject **) & gl_other_context, NULL);
     return NULL;
   }
 #else
@@ -1293,3 +1329,27 @@ gst_vaapi_plugin_base_get_allowed_raw_caps (GstVaapiPluginBase * plugin)
     return NULL;
   return plugin->allowed_raw_caps;
 }
+
+/**
+ * gst_vaapi_plugin_base_set_srcpad_can_dmabuf:
+ * @plugin: a #GstVaapiPluginBase
+ * @object: the GL context from gst-gl
+ *
+ * This function will determine if @object supports dmabuf
+ * importing.
+ *
+ * Please note that the context @object should come from downstream.
+ **/
+void
+gst_vaapi_plugin_base_set_srcpad_can_dmabuf (GstVaapiPluginBase * plugin,
+    GstObject * object)
+{
+#if VA_CHECK_VERSION (0,36,0) && USE_EGL && USE_GST_GL_HELPERS
+  GstGLContext *const gl_context = GST_GL_CONTEXT (object);
+
+  plugin->srcpad_can_dmabuf =
+      (!(gst_gl_context_get_gl_api (gl_context) & GST_GL_API_GLES1)
+      && gst_gl_context_check_feature (gl_context,
+          "EGL_EXT_image_dma_buf_import"));
+#endif
+}
