@@ -72,7 +72,8 @@ static const char gst_vaapipostproc_src_caps_str[] =
   GST_VAAPI_MAKE_GLTEXUPLOAD_CAPS "; "
 #endif
   GST_VIDEO_CAPS_MAKE (GST_VIDEO_FORMATS_ALL) ", "
-  GST_CAPS_INTERLACED_MODES;
+  GST_CAPS_INTERLACED_MODES "; "
+  GST_VAAPI_MAKE_DMABUF_CAPS;
 /* *INDENT-ON* */
 
 /* *INDENT-OFF* */
@@ -305,8 +306,10 @@ gst_vaapipostproc_start (GstBaseTransform * trans)
   ds_reset (&postproc->deinterlace_state);
   if (!gst_vaapi_plugin_base_open (GST_VAAPI_PLUGIN_BASE (postproc)))
     return FALSE;
-  if (!gst_vaapipostproc_ensure_filter (postproc))
-    return FALSE;
+  g_mutex_lock (&postproc->postproc_lock);
+  gst_vaapipostproc_ensure_filter (postproc);
+  g_mutex_unlock (&postproc->postproc_lock);
+
   return TRUE;
 }
 
@@ -1102,7 +1105,7 @@ ensure_allowed_srcpad_caps (GstVaapiPostproc * postproc)
 
 static GstCaps *
 gst_vaapipostproc_transform_caps_impl (GstBaseTransform * trans,
-    GstPadDirection direction, GstCaps * caps)
+    GstPadDirection direction)
 {
   GstVaapiPostproc *const postproc = GST_VAAPIPOSTPROC (trans);
 
@@ -1131,17 +1134,21 @@ gst_vaapipostproc_transform_caps (GstBaseTransform * trans,
       (direction == GST_PAD_SINK) ? "sink" : "src");
 
   g_mutex_lock (&postproc->postproc_lock);
-  caps = gst_vaapipostproc_transform_caps_impl (trans, direction, caps);
+  out_caps = gst_vaapipostproc_transform_caps_impl (trans, direction);
   g_mutex_unlock (&postproc->postproc_lock);
-  if (caps && filter) {
-    out_caps = gst_caps_intersect_full (caps, filter, GST_CAPS_INTERSECT_FIRST);
-    gst_caps_unref (caps);
-    caps = out_caps;
+
+  if (out_caps && filter) {
+    GstCaps *intersection;
+
+    intersection = gst_caps_intersect_full (out_caps, filter,
+        GST_CAPS_INTERSECT_FIRST);
+    gst_caps_unref (out_caps);
+    out_caps = intersection;
   }
 
-  GST_DEBUG_OBJECT (trans, "returning caps: %" GST_PTR_FORMAT, caps);
+  GST_DEBUG_OBJECT (trans, "returning caps: %" GST_PTR_FORMAT, out_caps);
 
-  return caps;
+  return out_caps;
 }
 
 static GstCaps *
@@ -1157,24 +1164,22 @@ gst_vaapipostproc_fixate_caps (GstBaseTransform * trans,
 
   if (direction == GST_PAD_SRC) {
     /* @TODO: we can do better */
-    othercaps = gst_caps_fixate (othercaps);
+    outcaps = gst_caps_fixate (othercaps);
     goto done;
   }
 
   g_mutex_lock (&postproc->postproc_lock);
-  if ((outcaps = gst_vaapipostproc_fixate_srccaps (postproc, caps, othercaps)))
-    gst_caps_replace (&othercaps, outcaps);
+  outcaps = gst_vaapipostproc_fixate_srccaps (postproc, caps, othercaps);
   g_mutex_unlock (&postproc->postproc_lock);
 
   /* set passthrough according to caps changes or filter changes */
   gst_vaapipostproc_set_passthrough (trans);
 
 done:
-  GST_DEBUG_OBJECT (trans, "fixated othercaps to %" GST_PTR_FORMAT, othercaps);
-  if (outcaps)
-    gst_caps_unref (outcaps);
+  GST_DEBUG_OBJECT (trans, "fixated othercaps to %" GST_PTR_FORMAT, outcaps);
+  gst_caps_unref (othercaps);
 
-  return othercaps;
+  return outcaps;
 }
 
 static gboolean
@@ -1878,8 +1883,12 @@ cb_channels_init (GstVaapiPostproc * postproc)
   if (postproc->cb_channels)
     return;
 
-  if (!gst_vaapipostproc_ensure_filter (postproc))
+  g_mutex_lock (&postproc->postproc_lock);
+  if (!gst_vaapipostproc_ensure_filter (postproc)) {
+    g_mutex_unlock (&postproc->postproc_lock);
     return;
+  }
+  g_mutex_unlock (&postproc->postproc_lock);
 
   filter_ops = postproc->filter_ops ? g_ptr_array_ref (postproc->filter_ops)
       : gst_vaapi_filter_get_operations (postproc->filter);
